package edu.georgetown.gucs.dictionary;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.Serializable;
import java.io.UnsupportedEncodingException;
import java.math.BigInteger;
import java.net.UnknownHostException;
import java.text.DecimalFormat;
import java.util.Comparator;
import java.util.Date;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Random;
import java.util.TreeMap;
import java.util.Vector;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.RejectedExecutionException;

import org.jdom.Document;
import org.jdom.Element;
import org.jdom.JDOMException;
import org.jdom.input.SAXBuilder;
import org.jdom.output.Format;
import org.jdom.output.XMLOutputter;

import edu.georgetown.gucs.tokenizers.TokenizerList;
import edu.georgetown.gucs.utility.BoundedExecutor;
import edu.georgetown.gucs.utility.ExceptionHandler;
import edu.georgetown.gucs.utility.FileLister;
import edu.georgetown.gucs.utility.Global;

/**
 * Creates a list of unique tokens extracted from a collection of documents that can be trimmed by removing tokens based
 * on various different attributes; used for creating fingerprints of documents that are based on words that appear in a
 * document collection
 * 
 * @author Clay Shields
 * @author Lindsay Neubauer
 * 
 **/
public class Dictionary implements Serializable, Iterable<Entry<String, DictionaryEntry>> {

  /**
   * Comparator that orders strings alphabetically ignoring case; sets the locale to English, Arabic, or Chinese
   **/
  private static class IgnoreCase implements Comparator<String>, Serializable {
    private static final long serialVersionUID = 5081186068508241364L;

    @Override
    public int compare(String one, String two) {
      return one.compareToIgnoreCase(two);
    }
  }

  // arbitrary serialVersionUID (since we implement Serializable)
  public static final long serialVersionUID = 42L;

  private String creatingProgram;
  private String creation;
  protected Map<String, DictionaryEntry> dictionary;
  private String dictionaryFileName;
  private String dictionaryName;
  private String directory;
  private String GUID;
  private String language;
  private double maxIDFfound;
  private int maxThread;
  protected int nextPosition;
  private String systemID;
  private boolean threaded;
  private TokenizerList tokenizers;
  protected int totalDocuments;
  private boolean trained;
  private boolean training;
  private boolean trimmed;
  private double trimmed_max;
  private double trimmed_min;
  private int version;

  /**
   * Constructor that sets default values for this dictionary. The default language is English.
   **/
  public Dictionary() {
    dictionaryName = "";
    dictionary = new TreeMap<String, DictionaryEntry>(new IgnoreCase());
    version = 0;
    nextPosition = 0;
    totalDocuments = 0;
    trained = false;
    maxIDFfound = 0.0;
    setCreatingProgram();
    setSystemID();
    GUID = new BigInteger(128, new Random()).toString();
    threaded = false;
    maxThread = Runtime.getRuntime().availableProcessors();
    trimmed = false;
    Global.monitor.writeToBuffer("Dictionary", "Constructor", "GUID=" + GUID);
    Global.monitor.writeToBuffer("Dictionary", "Constructor", "threaded=" + threaded);
    Global.monitor.writeToBuffer("Dictionary", "Constructor", "maxThread=" + maxThread);
    Global.monitor.writeToBuffer("Dictionary", "Constructor", "trimmed=" + trimmed);
  }

  /**
   * Constructor that loads this dictionary from a file
   * 
   * @param dictionaryFilename
   *          the filename containing a dictionary
   */
  public Dictionary(String dictionaryFilename) {
    Global.monitor.writeToBuffer("Dictionary", "Constructor", "filename=" + dictionaryFilename);
    this.loadDictionaryXML(dictionaryFilename);
  }

  /**
   * Loads a previously serialized XML dictionary object into this dictionary
   * 
   * @param filename
   *          the string filename of the XML dictionary file to load
   **/
  @SuppressWarnings("unchecked")
  public synchronized void loadDictionaryXML(String filename) {

    if (dictionary != null)
      dictionary.clear();
    dictionaryFileName = filename;

    try {
      File xml = new File(dictionaryFileName);
      SAXBuilder builder = new SAXBuilder();
      Document doc = (Document) builder.build(xml);
      Element dict = doc.getRootElement();

      dictionaryName = dict.getAttributeValue("name");
      GUID = dict.getAttributeValue("GUID");
      Global.monitor.writeToBuffer("Dictionary", "loadDictionaryXML", "GUID=" + GUID);
      systemID = dict.getAttributeValue("systemID");
      Global.monitor.writeToBuffer("Dictionary", "loadDictionaryXML", "systemID=" + systemID);
      version = Integer.valueOf(dict.getAttributeValue("version"));
      language = dict.getAttributeValue("language");
      creatingProgram = dict.getAttributeValue("creatingProgram");
      Global.monitor.writeToBuffer("Dictionary", "loadDictionaryXML", "creatingProgram=" + creatingProgram);
      creation = dict.getAttributeValue("creation");
      directory = dict.getAttributeValue("directory");
      Global.monitor.writeToBuffer("Dictionary", "loadDictionaryXML", "directory=" + directory);
      maxIDFfound = Double.valueOf(dict.getAttributeValue("maxIDF"));
      Global.monitor.writeToBuffer("Dictionary", "loadDictionaryXML", "maxIDFfound=" + maxIDFfound);
      totalDocuments = Integer.valueOf(dict.getAttributeValue("totalDocuments"));
      Global.monitor.writeToBuffer("Dictionary", "loadDictionaryXML", "totalDocuments=" + totalDocuments);
      trimmed = Boolean.valueOf(dict.getAttributeValue("trimmed"));
      Global.monitor.writeToBuffer("Dictionary", "loadDictionaryXML", "trimmed=" + trimmed);
      if (trimmed) {
        trimmed_min = Double.valueOf(dict.getAttributeValue("trimmed_min"));
        Global.monitor.writeToBuffer("Dictionary", "loadDictionaryXML", "trimmed_min=" + trimmed_min);
        trimmed_max = Double.valueOf(dict.getAttributeValue("trimmed_max"));
        Global.monitor.writeToBuffer("Dictionary", "loadDictionaryXML", "trimmed_max=" + trimmed_max);
      }
      trained = Boolean.valueOf(dict.getAttributeValue("trained"));
      Global.monitor.writeToBuffer("Dictionary", "loadDictionaryXML", "trained=" + trained);
      training = Boolean.valueOf(dict.getAttributeValue("training"));
      Global.monitor.writeToBuffer("Dictionary", "loadDictionaryXML", "training=" + training);
      nextPosition = Integer.valueOf(dict.getAttributeValue("nextPosition"));
      Global.monitor.writeToBuffer("Dictionary", "loadDictionaryXML", "nextPosition=" + nextPosition);

      Vector<String> tokenizerNames = new Vector<String>();
      for (Element tokenizerElement : (List<Element>) dict.getChildren("tokenizer")) {
        String tokenizer = tokenizerElement.getAttributeValue("name");
        if (tokenizerElement.getAttribute("parameter") != null) {
          tokenizer += " " + tokenizerElement.getAttributeValue("parameter");
        }
        tokenizerNames.add(tokenizer.trim());
      }
      setTokenizers(tokenizerNames);
      Global.monitor.writeToBuffer("Dictionary", "loadDictionaryXML", "tokenizers=" + tokenizers);

      for (Element tokenElement : (List<Element>) dict.getChildren("token")) {
        dictionary.put(
            tokenElement.getAttributeValue("name"),
            new DictionaryEntry(tokenElement.getAttributeValue("name"), Integer.valueOf(tokenElement
                .getAttributeValue("position")), Integer.valueOf(tokenElement.getAttributeValue("frequency")), Integer
                .valueOf(tokenElement.getAttributeValue("documentCount")), Double.valueOf(tokenElement
                .getAttributeValue("idf"))));
      }
    } catch (FileNotFoundException e) {
      ExceptionHandler.handleException(e);
    } catch (UnsupportedEncodingException e) {
      ExceptionHandler.handleException(e);
    } catch (JDOMException e) {
      ExceptionHandler.handleException(e);
    } catch (IOException e) {
      ExceptionHandler.handleException(e);
    }
  }

  /**
   * Reads an XML file that specifies which tokenizers to use for this dictionary
   * 
   * @param config_file
   *          the string name of the XML configuration file containing this dictionary's tokenizers
   **/
  @SuppressWarnings("unchecked")
  public void addTokenizers(String config_file) {
    Global.monitor.writeToBuffer("Dictionary", "addTokenizers", "config_file=" + config_file);
    try {
      File xml = new File(config_file);
      SAXBuilder builder = new SAXBuilder();
      Document doc = (Document) builder.build(xml);

      Element dict = doc.getRootElement();
      language = dict.getAttributeValue("language");
      Vector<String> tokenizerNames = new Vector<String>();
      for (Element tokenizerElement : (List<Element>) dict.getChildren("tokenizer")) {
        String tokenizer = tokenizerElement.getAttributeValue("name");
        if (tokenizerNames.size() == 0 && !tokenizer.contains("FileTokenizer")) {
          ExceptionHandler.handleException(new Exception("Dictionary: First tokenizer (" + tokenizer
              + ") must read from a file."));
        }

        if (tokenizerElement.getAttribute("parameter") != null) {
          tokenizer += " " + tokenizerElement.getAttributeValue("parameter");
        }
        tokenizerNames.add(tokenizer.trim());
      }
      setTokenizers(tokenizerNames);
      Global.monitor.writeToBuffer("Dictionary", "addTokenizers", "tokenizers=" + tokenizers);
    } catch (FileNotFoundException e) {
      ExceptionHandler.handleException(e);
    } catch (UnsupportedEncodingException e) {
      ExceptionHandler.handleException(e);
    } catch (JDOMException e) {
      ExceptionHandler.handleException(e);
    } catch (IOException e) {
      ExceptionHandler.handleException(e);
    }
  }

  /**
   * Adds a document to this dictionary. Requires this dictionary to be in training mode.
   * 
   * @param documentIterator
   *          the iterator over all tokens in the document
   **/
  public void addDocument(Iterator<String> documentIterator) {
    if (training) {
      // iterate over all terms and insert them into a temp hash map
      Map<String, DictionaryEntry> file_input = new TreeMap<String, DictionaryEntry>(new IgnoreCase());
      while (documentIterator.hasNext()) {
        String token = documentIterator.next();
        if (file_input.containsKey(token)) {
          file_input.get(token).incrementFrequencyCount();
        } else {
          DictionaryEntry toInsert = new DictionaryEntry(token);
          file_input.put(token, toInsert);
        }
      }
      mergeDictionary(file_input);
    } else {
      ExceptionHandler.handleException(new Exception("Not in training mode"));
    }
  }

  /**
   * Creates a deep copy of the given dictionary object and stores it in this dictionary
   * 
   * @param dictionary
   *          the dictionary object to copy
   */
  public void clone(Dictionary dictionary) {
    this.dictionary.clear();
    this.dictionary.putAll(dictionary.dictionary);
    this.directory = dictionary.directory;
    Global.monitor.writeToBuffer("Dictionary", "clone", "directory=" + directory);
    this.language = dictionary.language;
    Global.monitor.writeToBuffer("Dictionary", "clone", "language=" + language);
    this.nextPosition = dictionary.nextPosition;
    Global.monitor.writeToBuffer("Dictionary", "clone", "nextPosition=" + nextPosition);
    this.trained = dictionary.trained;
    Global.monitor.writeToBuffer("Dictionary", "clone", "trained=" + trained);
    this.training = dictionary.training;
    Global.monitor.writeToBuffer("Dictionary", "clone", "training=" + training);
    this.version = dictionary.version;
    Global.monitor.writeToBuffer("Dictionary", "clone", "version=" + version);
    this.trimmed = dictionary.trimmed;
    Global.monitor.writeToBuffer("Dictionary", "clone", "trimmed=" + trimmed);
    if (this.trimmed) {
      this.trimmed_min = dictionary.trimmed_min;
      Global.monitor.writeToBuffer("Dictionary", "clone", "trimmed_min=" + trimmed_min);
      this.trimmed_max = dictionary.trimmed_max;
      Global.monitor.writeToBuffer("Dictionary", "clone", "trimmed_max=" + trimmed_max);
    }
    this.maxIDFfound = dictionary.maxIDFfound;
    Global.monitor.writeToBuffer("Dictionary", "clone", "maxIDFfound=" + maxIDFfound);
    this.totalDocuments = dictionary.totalDocuments;
    Global.monitor.writeToBuffer("Dictionary", "clone", "totalDocuments=" + totalDocuments);
    this.tokenizers = dictionary.tokenizers;
    Global.monitor.writeToBuffer("Dictionary", "clone", "tokenizers=" + tokenizers);
  }

  /**
   * Determines if this dictionary contains a particular token
   * 
   * @param token
   *          the string containing the token value
   * 
   * @return the <code>boolean</code> value of whether the token is present in this dictionary
   **/
  public boolean containsToken(String token) {
    return dictionary.containsKey(token);
  }

  /**
   * Creates this dictionary using an optional number or percent of the files in the given fileLister; if no count or
   * percent is specified then uses entire fileList to create dictionary
   * 
   * @param fileList
   *          the fileLister containing the files to use for this dictionary
   */
  public void createFromFileLister(FileLister fileList) {
    dictionary.clear();
    directory = fileList.getDirectory();
    this.createFromIterator(fileList.iterator());
  }

  /**
   * Creates this dictionary using an optional number or percent of the files in the given fileLister
   * 
   * @param fileList
   *          the fileLister containing the files to use for this dictionary
   * @param parameter
   *          the <code>double</code> count (>0) or percent (<0) of the files in the fileLister to use to create this
   *          dictionary
   * 
   * @return the fileLister containing all the files NOT used to create this dictionary
   */
  public FileLister createFromFileLister(FileLister fileList, double parameter) {
    dictionary.clear();
    directory = fileList.getDirectory();

    if (parameter < 1) {
      this.createFromIterator(fileList.createSampleListFromPercent(parameter * 10));
    } else {
      this.createFromIterator(fileList.createSampleListFromCount((int) parameter));
    }

    return fileList;
  }

  /**
   * Creates this dictionary from the files contained in the given iterator
   * 
   * @param fileIterator
   *          the iterator over the files to use to create this dictionary
   **/
  private void createFromIterator(Iterator<File> fileIterator) {
    if (threaded) {
      createFromIteratorThreaded(fileIterator);
    } else {
      Global.monitor.writeToBuffer("Dictionary", "createFromIterator", "");
      startTraining(tokenizers.getNames());

      Global.monitor.writeToBuffer("Dictionary", "createFromIterator", "adding documents");
      while (fileIterator.hasNext()) {
        Iterator<String> tokens = tokenizers.tokenizeFile(fileIterator.next());
        if (tokens != null) {
          this.addDocument(tokens);
        }
      }
      Global.monitor.writeToBuffer("Dictionary", "createFromIterator", "documents added");
      endTraining();
    }
  }

  /**
   * Uses multiple threads to speed execution of creating this dictionary from the files contained in the given iterator
   * 
   * @param fileIterator
   *          the iterator over the files to use to create this dictionary
   **/
  private void createFromIteratorThreaded(Iterator<File> fileIterator) {
    Global.monitor.writeToBuffer("Dictionary", "createFromIteratorThreaded", "");
    try {
      // Put this dictionary into training mode
      startTraining(tokenizers.getNames());

      ExecutorService exec = Executors.newFixedThreadPool(maxThread);
      BoundedExecutor bex = new BoundedExecutor(exec, maxThread);

      while (fileIterator.hasNext()) {
        DictionaryWorker currentFileProcessor = new DictionaryWorker(fileIterator.next(), tokenizers.getNames(), this);
        bex.submitTask(currentFileProcessor);
      }

      exec.shutdown();

      while (!exec.isTerminated()) {
        // Wait a second and see if we are done
        Thread.sleep(1000);
      }
      endTraining();
    } catch (RejectedExecutionException e) {
      ExceptionHandler.handleException(e);
    } catch (InterruptedException e) {
      ExceptionHandler.handleException(e);
    }
  }

  /**
   * Allows for operations on the tokens in this dictionary
   * 
   * @return the iterator over the tokens in this dictionary
   **/
  @Override
  public Iterator<Entry<String, DictionaryEntry>> iterator() {
    return dictionary.entrySet().iterator();
  }

  /**
   * Starts training, which permits tokens to be added to this dictionary. Specifies the names of the tokenizers to use
   * on the new tokens.
   * 
   * @param tokenizerVec
   *          the string list containing tokenizer names; must match the tokenizers used to create this dictionary
   **/
  private void startTraining(List<String> tokenizerVec) {
    Global.monitor.writeToBuffer("Dictionary", "startTraining", "");
    if (trimmed) {
      ExceptionHandler.handleException(new Exception("Dictionary: Training a trimmed dictionary is a bad idea"));
    }
    setTokenizers(tokenizerVec);
    training = true;
  }

  /**
   * Ends training, which prohibits adding new tokens to this dictionary and allows this dictionary to be written to a
   * file. Increments the version number and calculates the largest IDF in this dictionary.
   **/
  private void endTraining() {
    if (training) {
      version++;
      trained = true;
      training = false;
      computeIDFs();
    } else {
      ExceptionHandler.handleException(new Exception("Not in training mode"));
    }
    Global.monitor.writeToBuffer("Dictionary", "endTraining", "");
  }
  

  /**
   * Computes the Inverse Document Frequencies (IDFs) for each token in the dictionary and determines the largest IDF
   **/
  private void computeIDFs() {
    for (Entry<String, DictionaryEntry> token : this) {
      double idf = token.getValue().computeIDF(totalDocuments);
      maxIDFfound = (idf > maxIDFfound) ? idf : maxIDFfound;
    }
    Global.monitor.writeToBuffer("Dictionary", "findMaxIDF", "maxIDFfound=" + maxIDFfound);
  }

  /**
   * Provides the program that created this dictionary
   * 
   * @return the string name of the program that created this dictionary
   **/
  public String getCreatingProgram() {
    return creatingProgram;
  }

  /**
   * Provides the original creation date of this dictionary
   * 
   * @return the string date when this dictionary was created
   **/
  public String getCreation() {
    return creation;
  }

  /**
   * Provides the filename of this dictionary, if this dictionary was loaded from or saved to a file
   * 
   * @return the string dictionary filename
   **/
  public synchronized String getDictionaryFilename() {
    return dictionaryFileName;
  }

  /**
   * Provides the name of this dictionary
   * 
   * @return the string name of this dictionary
   **/
  public String getDictionaryName() {
    return dictionaryName;
  }

  /**
   * Provides the unique identifier for this dictionary
   * 
   * @return the string GUID for dictionary
   **/
  public String getGUID() {
    return GUID;
  }

  /**
   * Provides the language of this dictionary; assumes English if no language was previously specified
   * 
   * @return the string language of the files used to create this dictionary
   */
  public String getLanguage() {
    return (language != null) ? language : "english";
  }

  /**
   * Provides the largest IDF value in this dictionary
   * 
   * @return the <code>double</code> value of the largest IDF in this dictionary
   */
  public double getMaxIDF() {
    return maxIDFfound;
  }

  /**
   * Provides the normalized inverse document frequency (IDF) of the given token in this dictionary. The normalized IDF
   * is the token's IDF divided by the highest IDF in this dictionary.
   * 
   * @param token
   *          the string containing the token value
   * 
   * @return the <code>double</code> value of the token's normalized inverse document frequency (IDF); returns 0.0 if
   *         the token is not present in this dictionary
   **/
  public synchronized double getNormalizedIDF(String token) {
    if (!training) {
      if (dictionary.containsKey(token)) {
        return getNormalizedIDF(dictionary.get(token));
      } else {
        ExceptionHandler.handleException(new Exception("Token not in dictionary"));
      }
    } else {
      ExceptionHandler.handleException(new Exception("Dictionary in training mode"));
    }
    return 0.0;
  }

  /**
   * Provides the normalized inverse document frequency (IDF) of the given dictionaryEntry. The normalized IDF is the
   * token's IDF divided by the highest IDF in this dictionary.
   * 
   * @param entry
   *          the dictionaryEntry to calculate the normalized IDF
   * 
   * @return the <code>double</code> value of the dictionaryEntry's normalized inverse document frequency (IDF)
   */
  public double getNormalizedIDF(DictionaryEntry entry) {
    return entry.getIDF() / maxIDFfound;    
  }

  /**
   * Provides the position of the given token in this dictionary
   * 
   * @param token
   *          the string containing the token value
   * 
   * @return the position of the token; returns -1 if the token is not present in this dictionary
   **/
  public int getPosition(String token) {

    if (!training) {
      if (dictionary.containsKey(token)) {
        return dictionary.get(token).getPosition();
      } 
    } else {
      ExceptionHandler.handleException(new Exception("Dictionary in training mode"));
    }
    return -1;
  }

  /**
   * Provides the number of tokens in this dictionary
   * 
   * @return the number of unique tokens in this dictionary
   **/
  public int getSize() {
    return dictionary.size();
  }

  /**
   * Provides the directory used for this dictionary
   * 
   * @return the string directory for which this dictionary was created
   **/
  public String getSource() {
    return directory;
  }

  /**
   * Provides the identifier for the system that this dictionary was created on
   * 
   * @return the string system ID for this dictionary
   **/
  public String getSystemID() {
    return systemID;
  }

  /**
   * Provides the tokenizerList object used to create this dictionary
   * 
   * @return the tokenizerList object used for this dictionary
   **/
  public TokenizerList getTokenizers() {
    return tokenizers;
  }

  /**
   * Provides a vector of the tokens in this dictionary
   * 
   * @return the list of tokens in this dictionary
   **/
  public List<String> getTokens() {
    return new Vector<String>(dictionary.keySet());
  }

  /**
   * Provides the number of documents processed for this dictionary
   * 
   * @return the total number of documents used to create this dictionary
   **/
  public synchronized int getTotalDocuments() {
    return totalDocuments;
  }

  /**
   * Provides the version number of this dictionary. The version is incremented when training on this dictionary ends.
   * 
   * @return the version number of this dictionary
   **/
  public int getVersion() {
    return version;
  }

  /**
   * Creates dictionary from the given directory; Reads an XML file that specifies which tokenizers to use and outputs
   * dictionary to a file
   * 
   * @param config_file
   *          the string name of the XML configuration file containing dictionary tokenizers
   * @param directory
   *          the string path to the directory used to create this dictionary
   * @param output_file
   *          the string name/path to write this dictionary to
   */
  public void makeDictionary(String config_file, String directory, String output_file) {
    if (output_file == null || config_file == null || directory == null) {
      ExceptionHandler.handleException(new Exception("Dictionary: Null parameter(s) given."));
    }

    addTokenizers(config_file);

    if (tokenizers.getNames() == null) {
      ExceptionHandler.handleException(new Exception("Dictionary: Null tokenizerList"));
    }
    if (tokenizers.getNames().size() < 1) {
      ExceptionHandler.handleException(new Exception("Dictionary: Empty tokenizerList"));
    }
    if (dictionary == null) {
      ExceptionHandler.handleException(new Exception("Dictionary: Null dictionary"));
    }

    threadingOn();

    FileLister files = new FileLister();
    files.recursiveLoadDirectory(directory);
    createFromFileLister(files);

    saveDictionaryXML(output_file);
  }

  /**
   * Merges the given Map of tokens to dictionaryEntry with this dictionary
   * 
   * @param documentTokens
   *          the map containing the dictionary information to merge with this dictionary
   */
  protected synchronized void mergeDictionary(Map<String, DictionaryEntry> documentTokens) {
    for (Entry<String, DictionaryEntry> documentEntry : documentTokens.entrySet()) {
      if (dictionary.containsKey(documentEntry.getKey())) {
        // If so, increment the document count and frequency count
        dictionary.get(documentEntry.getKey()).incrementFrequencyCount(documentEntry.getValue().getFrequencyCount());
        dictionary.get(documentEntry.getKey()).incrementDocumentCount();
      } else {
        // Otherwise, insert the DictionaryEntry into this dictionary
        documentEntry.getValue().setPosition(nextPosition);
        nextPosition++;
        dictionary.put(documentEntry.getKey(), documentEntry.getValue());
      }
    }
    totalDocuments++;
  }

  /**
   * Writes this dictionary as a serialized XML object. Requires dictionary to be out of training mode.
   * 
   * @param filename
   *          the string XML filename to write this dictionary to
   **/
  public void saveDictionaryXML(String filename) {
    Global.monitor.writeToBuffer("Dictionary", "saveDictionaryXML", "start");
    Global.monitor.writeToBuffer("Dictionary", "saveDictionaryXML", "threaded=" + threaded);
    Global.monitor.writeToBuffer("Dictionary", "saveDictionaryXML", "maxThread=" + maxThread);
    if (!training) {
      dictionaryFileName = (new File(filename)).getAbsolutePath();
      dictionaryName = filename.substring(filename.lastIndexOf('/') + 1);
      creation = String.valueOf(new Date());
      DecimalFormat rounding = new DecimalFormat("#0.####");

      Element dict = new Element("dictionary");
      dict.setAttribute("name", dictionaryFileName);
      dict.setAttribute("GUID", GUID);
      dict.setAttribute("systemID", systemID);
      dict.setAttribute("version", String.valueOf(version));
      dict.setAttribute("language", language);
      if (getCreatingProgram() != null && getCreatingProgram().length() > 0)
        dict.setAttribute("creatingProgram", creatingProgram);
      if (getCreation() != null && getCreation().length() > 0)
        dict.setAttribute("creation", creation);
      if (getSource() != null && getSource().length() > 0)
        dict.setAttribute("directory", directory);
      dict.setAttribute("maxIDF", rounding.format((maxIDFfound)));
      dict.setAttribute("totalDocuments", String.valueOf(totalDocuments));
      dict.setAttribute("trimmed", String.valueOf(trimmed));
      if (trimmed) {
        dict.setAttribute("trimmed_min", String.valueOf(trimmed_min));
        dict.setAttribute("trimmed_max", String.valueOf(trimmed_max));
      }
      dict.setAttribute("trained", String.valueOf(trained));
      dict.setAttribute("training", String.valueOf(training));
      dict.setAttribute("nextPosition", String.valueOf(nextPosition));

      for (String tokenizer : tokenizers.getNames()) {
        String[] split = tokenizer.split("\\s");
        Element tokenizerElement = new Element("tokenizer");
        tokenizerElement.setAttribute("name", split[0]);
        StringBuffer parameters = new StringBuffer();
        if (split.length > 1) {
          parameters.append(split[1]);
          tokenizerElement.setAttribute("parameter", split[1]);
        }
        dict.addContent(tokenizerElement);
      }

      for (Entry<String, DictionaryEntry> entry : this) {
        Element tokenElement = new Element("token");
        tokenElement.setAttribute("name", entry.getKey());
        tokenElement.setAttribute("position", String.valueOf(entry.getValue().getPosition()));
        tokenElement.setAttribute("frequency", String.valueOf(entry.getValue().getFrequencyCount()));
        tokenElement.setAttribute("documentCount", String.valueOf(entry.getValue().getDocumentCount()));
        tokenElement.setAttribute("idf", rounding.format((entry.getValue().getIDF())));
        tokenElement.setAttribute("idfNorm", rounding.format((getNormalizedIDF(entry.getValue()))));
        dict.addContent(tokenElement);
      }

      try {
        Document doc = new Document(dict);
        FileOutputStream dictionaryOut = new FileOutputStream(dictionaryFileName);
        Format form = Format.getPrettyFormat();
        form.setEncoding("UTF-8");
        XMLOutputter out = new XMLOutputter(form);
        out.output(doc, dictionaryOut);
        System.out.println("Writing dictionary to " + dictionaryFileName);
        dictionaryOut.close();
      } catch (FileNotFoundException e) {
        ExceptionHandler.handleException(e);
      } catch (IOException e) {
        ExceptionHandler.handleException(e);
      }
    } else {
      ExceptionHandler.handleException(new Exception("No saves in training mode."));
    }
    Global.monitor.writeToBuffer("Dictionary", "saveDictionaryXML", "end");
  }

  /**
   * Determines the name of the program that created this dictionary
   **/
  public void setCreatingProgram() {
    Global.monitor.writeToBuffer("Dictionary", "setCreatingProgram", "start");
    creatingProgram = "";
    Runtime runtime = Runtime.getRuntime();
    String cmds[] = { "jps", "-m" };
    Process proc;
    try {
      proc = runtime.exec(cmds);
      BufferedReader bufferedreader = new BufferedReader(new InputStreamReader(proc.getInputStream(), "UTF-8"));

      String line;
      while ((line = bufferedreader.readLine()) != null) {
        for (String output : line.split("\\s")) {
          if (output.contains(".jar")) {
            creatingProgram = output;
            Global.monitor.writeToBuffer("Dictionary", "setCreatingProgram", "creatingProgram=" + creatingProgram);
          }
        }
      }
      bufferedreader.close();
    } catch (IOException e) {
      ExceptionHandler.handleException(e);
    }
  }

  /**
   * Sets the name of the program that created this dictionary
   * 
   * @param creatingProgram
   *          the string program that created this dictionary
   */
  public void setCreatingProgram(String creatingProgram) {
    this.creatingProgram = creatingProgram;
    Global.monitor.writeToBuffer("Dictionary", "setCreatingProgram", "creatingProgram=" + creatingProgram);
  }

  /**
   * Sets the language of this dictionary
   * 
   * @param language
   *          the string language of the files used to create this dictionary
   */
  public void setLanguage(String language) {
    this.language = language;
    Global.monitor.writeToBuffer("Dictionary", "setLanguage", "language=" + this.language);
  }

  /**
   * Sets the system ID of this dictionary
   */
  private void setSystemID() {
    try {
      systemID = java.net.InetAddress.getLocalHost().getCanonicalHostName();
    } catch (UnknownHostException e) {
      ExceptionHandler.handleException(e);
    }
    Global.monitor.writeToBuffer("Dictionary", "setSystemID", "systemID=" + systemID);
  }

  /**
   * Loads the tokenizers to use for this dictionary from a TokenizerList object. Only works if no tokenizers have
   * already been set.
   * 
   * @param tokenizerList
   *          the TokenizerList object containing the tokenizers to use for this dictionary
   */
  public void setTokenizers(TokenizerList tokenizerList) {
    if (tokenizers == null) {
      tokenizers = tokenizerList;
    } else {
      ExceptionHandler.handleException(new Exception("Dictionary: Cannot reset tokenizer list in a dictionary"));
    }
    Global.monitor.writeToBuffer("Dictionary", "setTokenizers", "tokenizers=" + tokenizers);
  }

  /**
   * Sets the tokenizers used by this dictionary; Only works if no tokenizers have already been set or if the given list
   * contains the same tokenizers as those that have already been set
   * 
   * @param tokenizerVec
   *          the list of tokenizer names to use to create this dictionary
   **/
  public void setTokenizers(List<String> tokenizerVec) {
    if (tokenizers == null) {
      tokenizers = new TokenizerList(tokenizerVec);
    } else if (tokenizers.getNames().size() != tokenizerVec.size() || !tokenizerVec.containsAll(tokenizers.getNames())) {
      ExceptionHandler.handleException(new Exception("Dictionary: Cannot reset tokenizers in a dictionary"));
    }
    Global.monitor.writeToBuffer("Dictionary", "setTokenizersByName", "tokenizers=" + tokenizers);
  }

  /**
   * Prints information about this dictionary, including all tokens and their statistics.
   **/
  public synchronized void showDictionary() {
    showStatistics();
    showTokens();
  }

  /**
   * Prints this dictionary statistics, including:
   * <ul>
   * <li>Name</li>
   * <li>Language</li>
   * <li>Number of documents</li>
   * <li>Number of tokens</li>
   * <li>If this dictionary has been trimmed (including the IDF range, if trimmed)</li>
   * <li>Minimum and maximum IDF</li>
   * <li>List of tokenizers</li>
   * </ul>
   **/
  public void showStatistics() {
    System.out.println("Name: " + dictionaryFileName);
    System.out.println("Language: " + language);
    System.out.println("# Docs: " + totalDocuments);
    System.out.println("# Tokens: " + dictionary.size());
    if (trimmed) {
      System.out.println("Trimmed: " + trimmed + " (min: " + trimmed_min + ", max: " + trimmed_max + ")");
    } else {
      System.out.println("Trimmed: " + trimmed);
    }
    System.out.println("Max IDF: " + maxIDFfound);
    System.out.println("Tokenizers: " + tokenizers);
  }

  /**
   * Prints each token in this dictionary with it's frequency, IDF, and normalized IDF
   **/
  public void showTokens() {
    System.out.format("%-25s %-10s %-10s %-10s%n", "Token", "Freq", "IDF", "Norm. IDF");
    System.out.println("----------------------------------------------------------");
    for (Entry<String, DictionaryEntry> entry : this) {
      System.out.format("%-25s %-10s %-10.4f %-10.4f%n", entry.getKey(), entry.getValue().getFrequencyCount(), entry
          .getValue().getIDF(), getNormalizedIDF(entry.getValue()));
    }
  }

  /**
   * Turns off threading for creating this dictionary
   **/
  public void threadingOff() {
    threaded = false;
    Global.monitor.writeToBuffer("Dictionary", "threadingOff", "threaded=" + threaded);
  }

  /**
   * Turns on threading for creating this dictionary. This should only be used for stand-alone dictionary creation (not
   * inside a threaded application).
   **/
  public void threadingOn() {
    threaded = true;
    Global.monitor.writeToBuffer("Dictionary", "threadingOn", "threaded=" + threaded);
  }

  /**
   * Turns on threading for creating this dictionary. This should only be used for stand-alone dictionary creation (not
   * inside a threaded application).
   * 
   * @param numThreads
   *          the number of threads to use to create this dictionary
   **/
  public void threadingOn(int numThreads) {
    maxThread = numThreads;
    threaded = true;
    Global.monitor.writeToBuffer("Dictionary", "threadingOn", "threaded=" + threaded);
    Global.monitor.writeToBuffer("Dictionary", "threadingOn", "maxThread=" + maxThread);
  }

  /**
   * Trims this dictionary by removing any token that is outside a range of normalized IDFs. Once trimmed, no new tokens
   * can be added to this dictionary.
   * 
   * @param min
   *          the <code>double</code> minimum normalized IDF to keep; any token with a lower IDF will be discarded.
   * @param max
   *          the <code>double</code> maximum normalized IDF to keep; any token with a higher IDF will be discarded.
   **/
  public void trimByIDF(double min, double max) {
    Global.monitor.writeToBuffer("Dictionary", "trimByIDF", "start");

    // Trims this dictionary to contain only tokens within the specified range of normalized IDF values
    if (trimmed) {
      ExceptionHandler.handleException(new Exception(
          "Dictionary: trimming an already trimmed dictionary is probably a bad idea"));
    }

    trimmed = true;
    trimmed_min = min;
    trimmed_max = max;
    Global.monitor.writeToBuffer("Dictionary", "trimByIDF", "trimmed=" + trimmed);
    Global.monitor.writeToBuffer("Dictionary", "trimByIDF", "trimmed_min=" + trimmed_min);
    Global.monitor.writeToBuffer("Dictionary", "trimByIDF", "trimmed_max=" + trimmed_max);

    Map<String, DictionaryEntry> trimmedDictionary = new TreeMap<String, DictionaryEntry>(new IgnoreCase());
    int currentPlace = 0;
    for (Entry<String, DictionaryEntry> entry : this) {
      if ((getNormalizedIDF(entry.getValue()) >= min) && (getNormalizedIDF(entry.getValue()) <= max)) {
        entry.getValue().setPosition(currentPlace++);
        trimmedDictionary.put(entry.getKey(), entry.getValue());
      }
    }

    dictionary = trimmedDictionary;
    Global.monitor.writeToBuffer("Dictionary", "trimByIDF", "end");
  }

  /**
   * Creates a dictionary from the specified directory using the provided tokenizers
   * 
   * @param args
   *          array of string command line arguments<br>
   *          <code>args[0]</code> the filename of the tokenizer configuration file<br>
   *          <code>args[1]</code> the directory to create the dictionary from<br>
   *          <code>args[2]</code> the filename to write the dictionary to
   */
  public static void main(String args[]) {
    String output_file = null;
    String directory = null;
    String config_file = null;

    try {
      config_file = args[0];
      directory = args[1];
      output_file = args[2];
    } catch (ArrayIndexOutOfBoundsException e) {
      System.err
          .println("Usage: java CreateDictionary <tokenizers config file> <directory> <output dictionary filename> ");
      System.err.println("Recursively creates a dictionary from the specified directory using the provided tokenizers");
      System.exit(-1);
    }
    Global.monitor.writeToBuffer("Dictionary", "CreateDictionary", "config=" + config_file);
    Global.monitor.writeToBuffer("Dictionary", "CreateDictionary", "directory=" + directory);
    Global.monitor.writeToBuffer("Dictionary", "CreateDictionary", "output=" + output_file);

    Dictionary dict = new Dictionary();
    dict.makeDictionary(config_file, directory, output_file);
    Global.monitor.closeBuffer();
  }
}
