package edu.georgetown.gucs.experiment;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.Vector;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.RejectedExecutionException;

import org.jdom.Document;
import org.jdom.Element;
import org.jdom.JDOMException;
import org.jdom.input.SAXBuilder;

import edu.georgetown.gucs.dictionary.Dictionary;
import edu.georgetown.gucs.fingerprinter.Fingerprinter;
import edu.georgetown.gucs.utility.BoundedExecutor;
import edu.georgetown.gucs.utility.ExceptionHandler;
import edu.georgetown.gucs.utility.FileLister;
import edu.georgetown.gucs.utility.ObjectCloner;

/**
 * Runs many {@linkplain Trial}s to determine the best settings to use for {@linkplain Dictionary} and
 * {@linkplain Fingerprinter} creation. Setting results print to standard output and optionally written to a database.
 * 
 * @author Clay Shields
 * @author Lindsay Neubauer
 * 
 */
public class Experiment {

  private boolean insert;
  private int numberOfTrials;
  private int numberOfTrialThreads;
  private int completedTrials = 0;
  private String language;
  private String description;
  private String config_file;
  private String dataset;
  private File datasetPath;
  private FileLister datasetFiles;
  private DBInterface database;
  private List<Integer> completedExperimentIDs = new Vector<Integer>();
  private List<Long> randomSeeds = new Vector<Long>();
  private Set<String> plots = new HashSet<String>();
  private Map<List<String>, List<String>> tokenizers_manglers = new HashMap<List<String>, List<String>>();
  private List<String> fingerprinters = new Vector<String>();
  private List<Double> dictionaryParameters = new Vector<Double>();
  private List<Double> sampleParameters = new Vector<Double>();
  private List<Double> minIDFs = new Vector<Double>();
  private List<Double> maxIDFs = new Vector<Double>();
  private Map<String, List<Integer>> matcherHash = new HashMap<String, List<Integer>>();

  /**
   * map of trialParameters to a LinkedHashMap of manglerName to vector of results (precision/recall/fscore/TP/FP/TN/FN)
   */
  private Map<TrialParameters, LinkedHashMap<String, Vector<Double[]>>> experimentManglerResults = new LinkedHashMap<TrialParameters, LinkedHashMap<String, Vector<Double[]>>>();
  private List<Double> precisionVector = new Vector<Double>();
  private List<Double> recallVector = new Vector<Double>();
  private List<Double> fscoreVector = new Vector<Double>();
  private List<Double> dictionarySizeVector = new Vector<Double>();

  /**
   * Constructor that loads properties from XML configuration file
   * 
   * @param config_file
   *          the string XML filename that specifies the configurations to be used
   */
  public Experiment(String config_file) {
    loadXMLConfig(config_file);
  }

  /**
   * Load properties from XML configuration file
   * 
   * @param config_file
   *          the string name of the configuration file
   */
  @SuppressWarnings("unchecked")
  public void loadXMLConfig(String config_file) {
    try {
      this.config_file = config_file;
      File xml = new File(config_file);
      SAXBuilder builder = new SAXBuilder();
      Document doc = (Document) builder.build(xml);

      Element root = doc.getRootElement();
      String threads = root.getAttributeValue("threads");
      numberOfTrialThreads = (threads != null && threads.length() > 0) ? Integer.valueOf(threads)
          : Runtime.getRuntime().availableProcessors();
      String trials = root.getAttributeValue("trials");
      numberOfTrials = (trials == null || Integer.valueOf(trials) < 3) ? 3 : Integer.valueOf(trials);
      String dbInsert = root.getAttributeValue("DB_insert");
      insert = (dbInsert != null) ? Boolean.valueOf(dbInsert) : false;
      if ((description = root.getAttributeValue("description")) == null) {
        System.err.println("Experiment: WARNING: Null description");
      }

      for (Element seed : (List<Element>) root.getChildren("randomSeed")) {
        randomSeeds.add(Long.valueOf(seed.getAttributeValue("seed")));
      }
      if (randomSeeds.size() != 0 && numberOfTrials != randomSeeds.size()) {
        ExceptionHandler.handleException(new Exception(
            "Experiment: Number of trials does not match number of random seeds. " + "Trials: " + numberOfTrials
                + " seeds size: " + randomSeeds.size()));
      }

      Element data = root.getChild("dataset");
      dataset = data.getAttributeValue("name");
      if (dataset == null || dataset.length() == 0) {
        ExceptionHandler.handleException(new Exception("Experiment: No dataset name provided"));
      }

      String path = data.getAttributeValue("path");
      if (path == null || path.length() == 0) {
        ExceptionHandler.handleException(new Exception("Experiment: No path to dataset provided"));
      }
      datasetPath = new File(path);

      for (Element sampleParameter : (List<Element>) data.getChildren("sample")) {
        if (sampleParameter.getAttributeValue("count") != null) {
          double parameter = Double.valueOf(sampleParameter.getAttributeValue("count"));
          if (parameter < 0) {
            ExceptionHandler.handleException(new Exception("Experiment: Negative sample counts"));
          } else if (parameter > 5000) {
            System.err
                .println("Experiment: Sample counts greater than 5000 can take a really long time when run with many manglers");
          }
          sampleParameters.add(parameter);
        }
        if (sampleParameter.getAttributeValue("percent") != null) {
          double parameter = Double.valueOf(sampleParameter.getAttributeValue("percent"));
          if (parameter < 1 || parameter > 100) {
            ExceptionHandler.handleException(new Exception("Experiment: Sample percentage out of range"));
          }
          sampleParameters.add(parameter / 1000);
        }
      }
      if (sampleParameters == null || sampleParameters.size() == 0) {
        ExceptionHandler.handleException(new Exception("Experiment: No sample settings provided"));
      }

      Element dict = data.getChild("dictionary");
      String lang = dict.getAttributeValue("language");
      language = (lang != null) ? lang : "english";
      for (Element dictParameter : (List<Element>) dict.getChildren("create")) {
        if (dictParameter.getAttributeValue("count") != null) {
          double parameter = Double.valueOf(dictParameter.getAttributeValue("count"));
          if (parameter < 0) {
            ExceptionHandler.handleException(new Exception("Experiment: Negative dictionary counts"));
          } else if (parameter == 0) {
            ExceptionHandler.handleException(new Exception("Experiment: Zero dictionary counts"));
          }
          dictionaryParameters.add(parameter);

        }
        if (dictParameter.getAttributeValue("percent") != null) {
          double parameter = Double.valueOf(dictParameter.getAttributeValue("percent"));
          if (parameter < 1 || parameter > 100) {
            ExceptionHandler.handleException(new Exception("Experiment: Dictionary percentage out of range"));
          }
          dictionaryParameters.add(parameter / 1000);
        }
      }
      if (dictionaryParameters == null || dictionaryParameters.size() == 0) {
        ExceptionHandler.handleException(new Exception("Experiment: No dictionary settings provided"));
      }

      if (dict.getChild("minIDF") != null) {
        for (Element min_idf : (List<Element>) dict.getChildren("minIDF")) {
          minIDFs.add(Double.valueOf(min_idf.getAttributeValue("norm")));
        }
      } else {
        minIDFs.add(0.0);
      }

      if (dict.getChild("maxIDF") != null) {
        for (Element max_idf : (List<Element>) dict.getChildren("maxIDF")) {
          maxIDFs.add(Double.valueOf(max_idf.getAttributeValue("norm")));
        }
      } else {
        maxIDFs.add(1.0);
      }

      for (Element fingerprint : (List<Element>) root.getChildren("fingerprinter")) {
        fingerprinters.add(fingerprint.getAttributeValue("name"));
      }
      if (fingerprinters == null || fingerprinters.size() == 0) {
        ExceptionHandler.handleException(new Exception("Experiment: No fingerprinter provided"));
      }

      for (Element matcherElement : (List<Element>) root.getChildren("matcher")) {
        List<Integer> matcherParameter = new Vector<Integer>();
        for (Element parameterElement : (List<Element>) matcherElement.getChildren("parameter")) {
          matcherParameter.add(Integer.valueOf(parameterElement.getAttributeValue("score")));
        }
        // If there are no matcher parameters then add the default minimum score
        // (0) to the vector
        if (matcherParameter.size() == 0) {
          matcherParameter.add(0);
        }
        matcherHash.put(matcherElement.getAttributeValue("name"), matcherParameter);
      }
      if (matcherHash == null || matcherHash.isEmpty()) {
        ExceptionHandler.handleException(new Exception("Experiment: No matcher provided"));
      }

      for (Element tokenizerListElement : (List<Element>) root.getChildren("tokenizers")) {
        Vector<String> tokenizerNames = new Vector<String>();
        Vector<String> manglerLists = new Vector<String>();
        for (Element tokenizerElement : (List<Element>) tokenizerListElement.getChildren("tokenizer")) {
          String tokenizer = tokenizerElement.getAttributeValue("name");
          if (tokenizerNames.size() == 0 && !tokenizer.contains("FileTokenizer")) {
            ExceptionHandler.handleException(new Exception("Experiment: First tokenizer (" + tokenizer
                + ") must read from a file."));
          }

          if (tokenizerElement.getAttributeValue("parameter") != null) {
            tokenizer += " " + tokenizerElement.getAttributeValue("parameter");
          } else if (tokenizer.equalsIgnoreCase("FileManglerTokenizer")) {
            for (Element manglersListElement : (List<Element>) tokenizerElement.getChildren("manglers")) {
              StringBuffer mangler = new StringBuffer();
              for (Element manglerElement : (List<Element>) manglersListElement.getChildren("mangler")) {
                mangler.append(" " + manglerElement.getAttributeValue("name").trim());
                if (!manglerElement.getAttributeValue("name").equalsIgnoreCase("null")) {
                  mangler.append("-" + manglerElement.getAttributeValue("percent").trim());
                }
              }
              manglerLists.add(mangler.toString().trim());
            }
            manglerLists.trimToSize();
            if (manglerLists == null || manglerLists.size() == 0) {
              ExceptionHandler.handleException(new Exception("Experiment: No manglers provided"));
            }
          }
          tokenizerNames.add(tokenizer.trim());
        }
        tokenizerNames.trimToSize();
        this.tokenizers_manglers.put(tokenizerNames, manglerLists);
      }
      if (tokenizers_manglers == null || tokenizers_manglers.size() == 0) {
        ExceptionHandler.handleException(new Exception("Experiment: No tokenizers provided"));
      }
    } catch (FileNotFoundException e) {
      ExceptionHandler.handleException(e);
    } catch (UnsupportedEncodingException e) {
      ExceptionHandler.handleException(e);
    } catch (JDOMException e) {
      ExceptionHandler.handleException(e);
    } catch (IOException e) {
      ExceptionHandler.handleException(e);
    }
  } // end loadXMLConfig method

  /**
   * Prints out the properties for this experiment
   */
  public void showSettings() {
    int matcherCount = 0;
    int manglerCount = 0;
    System.out.println("Dataset: " + dataset);
    System.out.println("Data path: " + datasetPath.getAbsolutePath());
    System.out.println("Description: " + description);
    System.out.println("Language: " + language);
    System.out.println("Database Insert: " + insert);
    System.out.println("Number of threads: " + numberOfTrialThreads);
    System.out.println("Number of trials: " + numberOfTrials);
    System.out.println();
    if (randomSeeds.size() > 0) {
      System.out.println("Random Seeds:");
      for (Long seed : randomSeeds) {
        System.out.println("     " + seed.toString());
      }
      System.out.println();
    }
    System.out.println("Dictionary settings:");
    for (Double parameter : dictionaryParameters) {
      if (parameter < 1) {
        System.out.println(" Percents: " + (int) (parameter * 1000));
      } else {
        System.out.println(" File Count: " + parameter.intValue());
      }
    }
    System.out.println();
    System.out.println("Sample settings:");
    for (Double parameter : sampleParameters) {
      if (parameter < 1) {
        System.out.println(" Percents: " + (int) (parameter * 1000));
      } else {
        System.out.println(" File Count: " + parameter.intValue());
      }
    }
    System.out.println();
    System.out.print("Minimum IDFs: ");
    System.out.println(minIDFs);
    System.out.print("Maximum IDFs: ");
    System.out.println(maxIDFs);
    System.out.println();
    System.out.println("Tokenizer Lists:");
    for (List<String> tokenizerList : tokenizers_manglers.keySet()) {
      System.out.println(" Tokenizers: " + tokenizerList);
      if (tokenizerList.contains("FileManglerTokenizer")) {
        System.out.println("   Manglers: " + tokenizers_manglers.get(tokenizerList));
        manglerCount += tokenizers_manglers.get(tokenizerList).size();
      }
    }
    System.out.println();
    System.out.println("Fingerprinter: ");
    for (String fingerprinter : fingerprinters) {
      System.out.println("     " + fingerprinter);
    }
    System.out.println("Matcher: ");
    for (String matcher : matcherHash.keySet()) {
      List<Integer> parameters = matcherHash.get(matcher);
      matcherCount += parameters.size();
      System.out.println("     " + matcher);
      System.out.println("          parameters: " + parameters);
    }
    System.out.println();
    if (plots.size() > 0) {
      System.out.println("Plots:");
      for (String plot : plots) {
        System.out.println("     " + plot);
      }
      System.out.println();
    }
    int totalExperiments = numberOfTrials * dictionaryParameters.size() * sampleParameters.size() * minIDFs.size()
        * maxIDFs.size() * tokenizers_manglers.keySet().size() * manglerCount * fingerprinters.size()
        * matcherHash.keySet().size() * matcherCount;
    System.out.println("Total runs: " + totalExperiments);
  } // end showSettings method

  /**
   * Runs this experiment
   * 
   * @return the <code>int</code> database run number for this experiment; returns 0 if not inserting into the database
   */
  public int runExperiment() {
    Global.progress.start("Experiment start");
    if (insert) {
      // Connect to the correct database
      database = new DBInterface(dataset);
      Global.progress.update("Experiment:    Connected to " + dataset + " database");
    }

    Global.progress.update("Experiment:    Running Trials");
    int runNumber = runTrials();

    Global.progress.update("Experiment:    Printing Trial Results");
    printTrialResults();
    printExperimentResults();

    if (insert) {
      Global.progress.update("Experiment:    Computing and storing experiment results in database");
      for (Integer experimentID : completedExperimentIDs) {
        database.insertExperimentResults(experimentID, numberOfTrials);
      }
      database.finishRun(runNumber);
      database.close();
      database = null;
    }

    Global.progress.stop("Experiment complete");
    return runNumber;
  }

  /**
   * Runs trials for each setting
   * 
   * @return the <code>int</code> database run number for this experiment; returns 0 if not inserting into the database
   */
  public int runTrials() {
    int runNumber = (insert) ? database.startNewRun() : 0;

    // Get the listing of files from the dataset. This is slow, so we only want
    // to do it once
    datasetFiles = new FileLister();
    datasetFiles.recursiveLoadDirectory(datasetPath.getAbsolutePath());

    // Create a thread pool to use to run the trials
    ExecutorService exec = Executors.newFixedThreadPool(numberOfTrialThreads);
    BoundedExecutor bex = new BoundedExecutor(exec, numberOfTrialThreads);

    for (Double currentDictionaryParameter : dictionaryParameters) {

      for (Double currentSampleParameter : sampleParameters) {

        for (Double currentMinIDF : minIDFs) {

          for (Double currentMaxIDF : maxIDFs) {

            // Only run tests where minIDF is less than maxIDF
            if (currentMinIDF < currentMaxIDF) {

              for (List<String> currentTokenizers : tokenizers_manglers.keySet()) {

                for (String currentFingerprinterName : fingerprinters) {

                  for (String currentMatcherName : matcherHash.keySet()) {
                    List<Integer> matcherParameter = matcherHash.get(currentMatcherName);

                    for (Integer currentMatcherParameter : matcherParameter) {

                      for (String manglers : tokenizers_manglers.get(currentTokenizers)) {

                        int experimentID = -1;
                        if (insert) {
                          experimentID = database.newExperiment(runNumber, description, dataset,
                              datasetPath.getAbsolutePath(), config_file, currentFingerprinterName, currentMatcherName,
                              currentMatcherParameter, currentMinIDF, currentMaxIDF, currentSampleParameter,
                              currentDictionaryParameter, language);
                          database.saveTokenizers(experimentID, currentTokenizers);
                          completedExperimentIDs.add(experimentID);
                        }

                        // Run the trials
                        TrialParameters trialParameters = null;
                        LinkedHashMap<String, Vector<Double[]>> manglerResults = new LinkedHashMap<String, Vector<Double[]>>();
                        for (int i = 0; i < numberOfTrials; i++) {
                          // Create a random seed, if there isn't one
                          long currentSeed = (randomSeeds == null || randomSeeds.size() == 0) ? System
                              .currentTimeMillis() : randomSeeds.get(i);

                          FileLister trialFiles = (FileLister) (ObjectCloner.deepCopy(datasetFiles));

                          Trial trial = new Trial(experimentID, insert, database, dataset, trialFiles,
                              datasetPath.getAbsoluteFile(), language, currentSeed, currentDictionaryParameter,
                              currentSampleParameter, currentMinIDF, currentMaxIDF, currentTokenizers, manglers,
                              currentFingerprinterName, currentMatcherName, currentMatcherParameter);

                          try {
                            bex.submitTask(trial);
                          } catch (RejectedExecutionException e) {
                            ExceptionHandler.handleException(e);
                          }
                          Map<String, Double[]> trialManglerResults = trial.getManglerResults();
                          trialParameters = trial.getTrialParameters();
                          completedTrials++;
                          dictionarySizeVector.add((double) trial.getDictionarySize());

                          for (Entry<String, Double[]> entry : trialManglerResults.entrySet()) {
                            Double[] runResult = entry.getValue();
                        
                            // Record all precision, recall, and f-score for
                            // total averages (over the experiment)
                            precisionVector.add(runResult[Global.PRECISION]);
                            recallVector.add(runResult[Global.RECALL]);
                            fscoreVector.add(runResult[Global.FSCORE]);
                            
                            Vector<Double[]> trialResults = (i == 0) ? new Vector<Double[]>() : manglerResults
                                .get(entry.getKey());
                            trialResults.add(runResult);
                            manglerResults.put(entry.getKey(), trialResults);
                          }
                        }
                        Global.progress.update("");
                        experimentManglerResults.put(trialParameters, manglerResults);
                      }
                    }
                  }
                }
              }
            }
          }
        }
      }
    }

    exec.shutdown();

    while (!exec.isTerminated()) {
      // Wait a second and see if we are done
      try {
        Thread.sleep(1000);
      } catch (InterruptedException e) {
        ExceptionHandler.handleException(e);
      }
    }
    return runNumber;
  } // runTrial

  /**
   * Calculates the results of all trials, grouped by parameters and displayed by mangler
   */
  public void printTrialResults() {
    System.out.println();
    System.out.println("**************");
    System.out.println("Trial Results");
    System.out.println("**************");
    for (TrialParameters trialParameters : experimentManglerResults.keySet()) {
      System.out.println(trialParameters);
      System.out.println();
      for (String mangler : experimentManglerResults.get(trialParameters).keySet()) {
        System.out.format(" %-15s %s%n", "Mangler(s):", mangler);
        double[] resultSum = new double[] { 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0 };
        for (Double[] manglerResults : experimentManglerResults.get(trialParameters).get(mangler)) {
          // Add results to sum array of result type (to be used to find average over trials)
          resultSum[Global.PRECISION] += manglerResults[Global.PRECISION];
          resultSum[Global.RECALL] += manglerResults[Global.RECALL];
          resultSum[Global.FSCORE] += manglerResults[Global.FSCORE];
          resultSum[Global.TRUE_POSITIVE] += manglerResults[Global.TRUE_POSITIVE];
          resultSum[Global.FALSE_POSITIVE] += manglerResults[Global.FALSE_POSITIVE];
          resultSum[Global.TRUE_NEGATIVE] += manglerResults[Global.TRUE_NEGATIVE];
          resultSum[Global.FALSE_NEGATIVE] += manglerResults[Global.FALSE_NEGATIVE];

          // Print mangler results by result type
          System.out.format("%13s: %-15.6f", "Precision", manglerResults[Global.PRECISION]);
          System.out.format("%13s: %-15.6f", "Recall", manglerResults[Global.RECALL]);
          System.out.format("%13s: %-15.6f", "F-Score", manglerResults[Global.FSCORE]);
          System.out.format("%6s: %-15d", "TP", manglerResults[Global.TRUE_POSITIVE].intValue());
          System.out.format("%6s: %-15d", "FP", manglerResults[Global.FALSE_POSITIVE].intValue());
          System.out.format("%6s: %-15d", "TN", manglerResults[Global.TRUE_NEGATIVE].intValue());
          System.out.format("%6s: %-15d", "FN", manglerResults[Global.FALSE_NEGATIVE].intValue());
          System.out.println();
        }

        // Calculate averages for each result type
        int numManglerResults = experimentManglerResults.get(trialParameters).get(mangler).size();
        Double[] averages = new Double[] { resultSum[Global.PRECISION] / numManglerResults,
            resultSum[Global.RECALL] / numManglerResults, resultSum[Global.FSCORE] / numManglerResults,
            resultSum[Global.TRUE_POSITIVE] / numManglerResults, resultSum[Global.FALSE_POSITIVE] / numManglerResults,
            resultSum[Global.TRUE_NEGATIVE] / numManglerResults, resultSum[Global.FALSE_NEGATIVE] / numManglerResults, };
        Vector<Double[]> updatedVector = experimentManglerResults.get(trialParameters).get(mangler);
        updatedVector.add(0, averages);
        experimentManglerResults.get(trialParameters).put(mangler, updatedVector);

        // Print mangler result averages by result type
        System.out.format("%13s: %-15.6f", "AvgPrecision", averages[Global.PRECISION]);
        System.out.format("%13s: %-15.6f", "AvgRecall", averages[Global.RECALL]);
        System.out.format("%13s: %-15.6f", "AvgF-Score", averages[Global.FSCORE]);
        System.out.format("%6s: %-15.6f", "AvgTP", averages[Global.TRUE_POSITIVE]);
        System.out.format("%6s: %-15.6f", "AvgFP", averages[Global.FALSE_POSITIVE]);
        System.out.format("%6s: %-15.6f", "AvgTN", averages[Global.TRUE_NEGATIVE]);
        System.out.format("%6s: %-15.6f", "AvgFN", averages[Global.FALSE_NEGATIVE]);
        System.out.println();
        System.out.println();
      }
    }
  }

  /**
   * Prints the results of all trials, grouped by parameters and displayed by mangler
   */
  public void printExperimentResults() {
    System.out.println("*******************");
    System.out.println("Experiment Results");
    System.out.println("*******************");
    final double tValue = Statistics.getTValueNinetyFive(completedTrials - 1);

    System.out.println("Precision:");
    double mean = Statistics.getMean(precisionVector);
    double stdDev = Statistics.getStandardDeviation(precisionVector, mean);
    System.out.format("  %-20s %-15.6f%n", "Average", mean);
    System.out.format("  %-20s %-15.6f%n", "Standard Deviation", stdDev);
    System.out.format("  %-20s %-15.6f%n", "Error (95 bound)", (stdDev / Math.sqrt(completedTrials)) * tValue);
    System.out.println("");

    System.out.println("Recall:");
    mean = Statistics.getMean(recallVector);
    stdDev = Statistics.getStandardDeviation(recallVector, mean);
    System.out.format("  %-20s %-15.6f%n", "Average", mean);
    System.out.format("  %-20s %-15.6f%n", "Standard Deviation", stdDev);
    System.out.format("  %-20s %-15.6f%n", "Error (95 bound)", (stdDev / Math.sqrt(completedTrials)) * tValue);
    System.out.println("");

    System.out.println("F-Score:");
    mean = Statistics.getMean(fscoreVector);
    stdDev = Statistics.getStandardDeviation(fscoreVector, mean);
    System.out.format("  %-20s %-15.6f%n", "Average", mean);
    System.out.format("  %-20s %-15.6f%n", "Standard Deviation", stdDev);
    System.out.format("  %-20s %-15.6f%n", "Error (95 bound)", (stdDev / Math.sqrt(completedTrials)) * tValue);
    System.out.println("");

    System.out.println("Dictionary Size:");
    mean = Statistics.getMean(dictionarySizeVector);
    stdDev = Statistics.getStandardDeviation(dictionarySizeVector, mean);
    System.out.format("  %-20s %-15.6f%n", "Average", mean);
    System.out.format("  %-20s %-15.6f%n", "Standard Deviation", stdDev);
    System.out.format("  %-20s %-15.6f%n", "Error (95 bound)", (stdDev / Math.sqrt(completedTrials)) * tValue);
    System.out.println("");
  }

  /**
   * Runs experiments using multiple runs with different settings and generates results
   * 
   * @param args
   *          array of string command line arguments<br>
   *          <code>args[0]</code> the filename of the experiment configuration file<br>
   *          <code>args[1]</code> the option to print ("show") results to the screen or insert ("insert") results into
   *          the database; if "insert" is chosen results also print to screen
   */
  public static void main(String args[]) {
    String config_file = null;
    try {
      config_file = args[0];
    } catch (ArrayIndexOutOfBoundsException e) {
      System.err.println("Usage: java Experiment <config_file>");
      System.exit(-1);
    }
    Experiment exp = new Experiment(config_file);
    exp.showSettings();
    exp.runExperiment();
  }
}
