package edu.georgetown.gucs.fingerprinter;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.UnsupportedEncodingException;
import java.math.BigInteger;
import java.net.UnknownHostException;
import java.util.Date;
import java.util.List;
import java.util.Random;
import java.util.Vector;

import org.apache.xerces.impl.dv.util.Base64;
import org.jdom.Document;
import org.jdom.Element;
import org.jdom.JDOMException;
import org.jdom.input.SAXBuilder;
import org.jdom.output.Format;
import org.jdom.output.XMLOutputter;

import edu.georgetown.gucs.dictionary.Dictionary;
import edu.georgetown.gucs.tokenizers.TokenizerList;
import edu.georgetown.gucs.utility.ExceptionHandler;

/**
 * This is the base class for all fingerprinters. A Fingerprinter tokenizes a document and compares it to the tokens in
 * a {@linkplain Dictionary} to create a fingerprint that can be used for recognition of the document or comparisons to
 * other documents and fingerprints. <br>
 * For experimental purposes, the fingerprinter controls a <code>mangler</code> which is used to add errors or noise to
 * the files being processed. This allows modification of the input dictionary on the fly, so that fingerprinters can be
 * tested for accuracy over errors.
 * 
 * @author Clay Shields
 **/
public class Fingerprinter {

  protected Dictionary dictionary;
  protected TokenizerList tokenizers;
  protected String base64Fingerprint;
  protected String fingerprintName;
  protected String GUID;
  protected String systemID;
  protected String creatingProgram;
  protected String creator;
  protected String targetFile;
  protected int version;
  protected String volume;
  protected String diskImage;
  protected String byteRun;
  protected List<String> unknownTokens;
  protected boolean showDigest;
  protected boolean showDataSource;
  protected boolean showDictionary;

  /**
   * Constructor that generates the fingerprint name, version, unique identifier (GUID), system identifier, and creating
   * program. Sets the fingerprinter output to display all available information.
   **/
  public Fingerprinter() {
    dictionary = null;
    tokenizers = null;

    Class<? extends Fingerprinter> classtype = this.getClass();
    String classname = classtype.getName();
    fingerprintName = classname;
    GUID = new BigInteger(128, new Random()).toString();
    version = 1;
    creator = "";
    creatingProgram = generateCreatingProgram();
    try {
      systemID = java.net.InetAddress.getLocalHost().getCanonicalHostName();
    } catch (UnknownHostException e) {
      ExceptionHandler.handleException(e);
    }

    setVerbose();
  }

  /**
   * Constructor that loads a dictionary and its tokenizers. Generates the fingerprint name, version, unique identifier
   * (GUID), system identifier, and creating program. Sets the fingerprinter output to display all available
   * information.
   * 
   * @param dict
   *          the dictionary to use to create this fingerprinter
   */
  public Fingerprinter(Dictionary dict) {
    dictionary = dict;
    tokenizers = new TokenizerList(dictionary.getTokenizerNames());

    Class<? extends Fingerprinter> classtype = this.getClass();
    String classname = classtype.getName();
    fingerprintName = classname;
    GUID = new BigInteger(128, new Random()).toString();
    version = 1;
    creator = "";
    creatingProgram = generateCreatingProgram();
    try {
      systemID = java.net.InetAddress.getLocalHost().getCanonicalHostName();
    } catch (UnknownHostException e) {
      ExceptionHandler.handleException(e);
    }

    setVerbose();
  }

  /**
   * Constructor that loads a dictionary and its tokenizers. Generates the fingerprint name, version, unique identifier
   * (GUID), system identifier, and creating program. Sets the fingerprinter output to display all available
   * information.
   * 
   * @param dictionaryFilename
   *          the filename containing the dictionary to use for this fingerprinter
   **/
  public Fingerprinter(String dictionaryFilename) {
    dictionary = new Dictionary(dictionaryFilename);
    tokenizers = new TokenizerList(dictionary.getTokenizerNames());

    Class<? extends Fingerprinter> classtype = this.getClass();
    String classname = classtype.getName();
    fingerprintName = classname;
    GUID = new BigInteger(128, new Random()).toString();
    version = 1;
    creator = "";
    creatingProgram = generateCreatingProgram();
    try {
      systemID = java.net.InetAddress.getLocalHost().getCanonicalHostName();
    } catch (UnknownHostException e) {
      ExceptionHandler.handleException(e);
    }

    setVerbose();
  }

  /**
   * Specifies which information to display in this fingerprinter's digest XML output
   * 
   * @param showDigest
   *          the boolean value specifying whether to display digest information
   * @param showDataSource
   *          the boolean value specifying whether to display data source information
   * @param showDictionary
   *          the boolean value specifying whether to display dictionary information
   **/
  public void setOutput(boolean showDigest, boolean showDictionary, boolean showDataSource) {
    this.showDigest = showDigest;
    this.showDataSource = showDataSource;
    this.showDictionary = showDictionary;
  }

  /**
   * Sets this fingerprinter's digest XML output to only display file and fingerprint information
   **/
  public void setTerse() {
    setOutput(false, false, false);
  }

  /**
   * Sets this fingerprinter's digest XML output to display all available information
   **/
  public void setVerbose() {
    // TODO: Change showDataSource to true once data source information is fully
    // implemented
    setOutput(true, true, false);
  }

  /**
   * Sets the creator (person/organization) of this fingerprinter. By default this value is an empty string.
   **/
  public void setCreator(String creator) {
    this.creator = creator;
  }

  /**
   * Passes the specified mangler settings to the mangler for this fingerprinter. The mangler is disabled if "null" is
   * passed as the settings.
   * 
   * @param settings
   *          the string of mangler settings for this fingerprinter
   */
  public void setMangler(String settings) {
    if (settings.contains("null")) {
      tokenizers.disableMangler();
    } else {
      tokenizers.enableMangler(settings);
    }
  }

  /**
   * Passes the specified mangler settings and a set of tokens to the mangler for this fingerprinter. The mangler is
   * disabled if "null" is passed as the settings.
   * 
   * @param settings
   *          the string of mangler settings for this fingerprinter
   * @param dict
   *          the dictionary to use to pass dictionary tokens to the mangler
   */
  public void setMangler(String settings, Dictionary dict) {
    if (dict == null) {
      System.err.println("Fingerprinter: Null dictionary");
    }

    if (settings.contains("null")) {
      tokenizers.disableMangler();
    } else {
      tokenizers.enableMangler(settings, dict.getTokens());
    }
  }

  /**
   * Sets a random number generator for the mangler to allow for repeatability. Repeatable behavior is sometimes used
   * when running experiments.
   * 
   * @param random
   *          the random object to use for this dictionary
   **/
  public void setManglerRNG(Random random) {
    tokenizers.setManglerRNG(random);
  }

  /**
   * Gives this fingerprinter's name. This is usually determined by getting the Class name.
   * 
   * @return the string name of this fingerprinter
   **/
  public String getFingerprintName() {
    return fingerprintName.substring(fingerprintName.lastIndexOf('.') + 1);
  }

  /**
   * Gives this fingerprinter's dictionary
   * 
   * @return the dictionary used for this fingerprinter
   **/
  public Dictionary getDictionary() {
    return dictionary;
  }

  /**
   * Gives this fingerprinter's fingerprint in Base64 encoding
   * 
   * @return the Base64 encoded string containing the fingerprint
   **/
  public String getBase64Fingerprint() {
    return base64Fingerprint;
  }

  /**
   * Gives the program that created this fingerprinter
   * 
   * @return the string of the creating program for this fingerprinter
   **/
  public String getCreatingProgram() {
    return creatingProgram;
  }

  /**
   * Determines the program that created this fingerprinter
   * 
   * @return the string of the creating program or null if none can be determined
   **/
  public String generateCreatingProgram() {
    creatingProgram = "";
    Runtime runtime = Runtime.getRuntime();
    String cmds[] = { "jps", "-m" };
    Process proc;
    try {
      proc = runtime.exec(cmds);
      InputStreamReader in = new InputStreamReader(proc.getInputStream(), "UTF-8");
      BufferedReader bufferedreader = new BufferedReader(in);

      String line;
      while ((line = bufferedreader.readLine()) != null) {
        for (String output : line.split("\\s")) {
          if (output.contains(".jar")) {
            bufferedreader.close();
            in.close();
            return output;
          }
        }
      }
      bufferedreader.close();
      in.close();
    } catch (IOException e) {
      ExceptionHandler.handleException(e);
    }
    return null;
  }

  /**
   * Computes the fingerprint of this document as a byte array; indicates the presence or absence of each token in this
   * dictionary
   * 
   * @param filename
   *          the string filename of the document to fingerprint
   * 
   * @return the byte array containing the fingerprint
   */
  public byte[] computeFingerprint(String filename) {
    return new byte[] {};
  }

  /**
   * Computes the fingerprint of a document as a Base64 encoded string; indicates the presence or absence of each token
   * in this dictionary
   * 
   * @param filename
   *          the string filename of the document to fingerprint
   * 
   * @return the Base64 encoded string containing the fingerprint
   */
  public String computeFingerprintXML(String filename) {
    return Base64.encode(computeFingerprint(filename));
  }

  /**
   * Sets the dictionary for this fingerprinter. This dictionary's tokens are compared against the tokenized document
   * that is being fingerprinted.
   * 
   * @param dict
   *          the dictionary to use for this fingerprinter
   **/
  public void setDictionary(Dictionary dict) {
    dictionary = dict;
    tokenizers = new TokenizerList(dictionary.getTokenizerNames());
    Class<? extends Fingerprinter> classtype = this.getClass();
    String classname = classtype.getName();
    fingerprintName = classname;
  }

  /**
   * Sets the dictionary for this fingerprinter. This dictionary's tokens are compared against the tokenized document
   * that is being fingerprinted.
   * 
   * @param filename
   *          the string filename of the dictionary to use for this fingerprinter
   **/
  public void setDictionary(String filename) {
    dictionary = new Dictionary(filename);
    tokenizers = new TokenizerList(dictionary.getTokenizerNames());
    Class<? extends Fingerprinter> classtype = this.getClass();
    String classname = classtype.getName();
    fingerprintName = classname;
  }

  /**
   * Loads a previously generated fingerprinter digest from a fingerprint XML file. <br>
   * Assumes the full dictionary is contained in the fingerprinter XML file or the file containing the dictionary
   * specified in the fingerprinter XML file exists in the current directory. If the full dictionary is contained in the
   * fingerprinter XML file but does not exist in the current directory, a dictionary XML file containing the dictionary
   * is created.
   * 
   * @param fingerprintFilename
   *          the string name of the XML file containing the fingerprinter's digest
   **/
  public void loadFingerprintXML(String fingerprintFilename) {
    loadFingerprintXML(fingerprintFilename, null);
  }

  /**
   * Loads a previously generated fingerprinter digest from fingerprinter and dictionary XML files. <br>
   * If dictionaryFilename is null, assumes the full dictionary is contained in the fingerprinter XML file or the file
   * containing the dictionary specified in the fingerprinter XML file exists in the current directory. If the full
   * dictionary is contained in the fingerprinter XML file but does not exist in the current directory, a dictionary XML
   * file containing the dictionary is created.
   * 
   * @param fingerprintFilename
   *          the string name of the XML file containing the fingerprinter's digest
   * @param dictionaryFilename
   *          the name of the XML file containing the dictionary
   **/
  @SuppressWarnings("unchecked")
  public void loadFingerprintXML(String fingerprintFilename, String dictionaryFilename) {
    FileOutputStream digestOut = null;
    try {
      File xml = new File(fingerprintFilename);
      SAXBuilder builder = new SAXBuilder();
      Document doc = (Document) builder.build(xml);
      Element fingerprinter = doc.getRootElement();

      fingerprintName = fingerprinter.getAttributeValue("fingerprinter");
      GUID = fingerprinter.getAttributeValue("GUID");
      version = Integer.valueOf(fingerprinter.getAttributeValue("version"));
      systemID = fingerprinter.getAttributeValue("systemID");
      creatingProgram = fingerprinter.getAttributeValue("creatingProgram");
      creator = fingerprinter.getAttributeValue("creator");

      Element fingerprint = fingerprinter.getChild("base64Fingerprint");
      base64Fingerprint = fingerprint.getTextTrim();

      showDataSource = !fingerprinter.getChildren("dataSource").isEmpty();
      if (showDataSource) {
        Element data = fingerprinter.getChild("dataSource");
        targetFile = data.getAttributeValue("filename");
        volume = data.getAttributeValue("volume");
        diskImage = data.getAttributeValue("diskImageID");
        byteRun = data.getAttributeValue("byteRun");
      }

      showDigest = !fingerprinter.getChildren("digest").isEmpty();
      if (showDigest) {
        Element digest = fingerprinter.getChild("digest");
        if (!digest.getChildren("unknownTokens").isEmpty()) {
          unknownTokens = new Vector<String>();
          for (Element unknownToken : (List<Element>) digest.getChild("unknownTokens").getChildren()) {
            unknownTokens.add(unknownToken.getAttributeValue("name"));
          }
        }
      }

      Element dict = fingerprinter.getChild("dictionary");
      showDictionary = !dict.getChildren().isEmpty();
      String filename = dict.getAttributeValue("name");
      File dictionaryCheck = new File(filename);
      if (dictionaryFilename != null) {
        dictionary = new Dictionary(dictionaryFilename);
        tokenizers = new TokenizerList(dictionary.getTokenizerNames());
        dictionary.setTokenizersByName(dictionary.getTokenizerNames());
        dictionary.setCreatingProgram(generateCreatingProgram());
      } else if (showDictionary && !dictionaryCheck.exists()) {
        Document document = new Document((Element) dict.clone());
        Format form = Format.getPrettyFormat();
        form.setEncoding("UTF-8");
        XMLOutputter out = new XMLOutputter(form);
        digestOut = new FileOutputStream(filename);
        out.output(document, digestOut);
        digestOut.close();
        dictionary = new Dictionary(filename);
        tokenizers = new TokenizerList();
        for (Element tokenizer : (List<Element>) dict.getChildren("tokenizer")) {
          tokenizers.addTokenizer(tokenizer.getTextTrim());
        }
        dictionary.setTokenizersByName(tokenizers.getNames());
        dictionary.setCreatingProgram(generateCreatingProgram());
      } else {
        dictionary = new Dictionary(filename);
        tokenizers = new TokenizerList(dictionary.getTokenizerNames());
        dictionary.setTokenizersByName(tokenizers.getNames());
        dictionary.setCreatingProgram(generateCreatingProgram());
      }
    } catch (FileNotFoundException e) {
      ExceptionHandler.handleException(e);
    } catch (UnsupportedEncodingException e) {
      ExceptionHandler.handleException(e);
    } catch (JDOMException e) {
      ExceptionHandler.handleException(e);
    } catch (IOException e) {
      ExceptionHandler.handleException(e);
    } finally {
      if (digestOut != null)
        try {
          digestOut.close();
        } catch (IOException e) {
          ExceptionHandler.handleException(e);
        }
    }
  }

  /**
   * Generates an XML file that contains this fingerprinter's digest
   * 
   * @param base64Fingerprint
   *          the Base64 encoded string containing the fingerprint
   * @param document
   *          the string filename of the document this fingerprint is for
   * @param outputFile
   *          the string XML filename to write the digest to
   */
  public void generateXML(String base64Fingerprint, String document, String outputFile) {
    this.base64Fingerprint = base64Fingerprint;
    targetFile = document;

    System.out.println("generating digest XML: " + (new File(outputFile)).getAbsolutePath());
    Element fingerprint = new Element("fingerprint");
    // Metadata
    fingerprint.setAttribute("fingerprinter", this.getClass().getSimpleName());
    fingerprint.setAttribute("GUID", GUID);
    fingerprint.setAttribute("version", String.valueOf(version));
    fingerprint.setAttribute("systemID", systemID);
    if (creator != null && creator.length() > 0)
      fingerprint.setAttribute("creator", creator);
    if (creatingProgram != null && creatingProgram.length() > 0)
      fingerprint.setAttribute("creatingProgram", creatingProgram);
    fingerprint.setAttribute("creation", String.valueOf(new Date()));

    Element base64Finger = new Element("base64Fingerprint");
    base64Finger.setText(base64Fingerprint);
    fingerprint.addContent(base64Finger);

    // Data source
    if (showDataSource) {
      Element dataSource = new Element("dataSource");
      dataSource.setAttribute("filename", (new File(targetFile)).getAbsolutePath());
      dataSource.setAttribute("filepath", dictionary.getSource());
      dataSource.setAttribute("systemID", dictionary.getSystemID());
      if (volume != null)
        dataSource.setAttribute("volume", volume);
      if (diskImage != null)
        dataSource.setAttribute("diskImageID", diskImage);
      if (byteRun != null)
        dataSource.setAttribute("byteRun", byteRun);
      fingerprint.addContent(dataSource);
    }

    // Digest
    if (showDigest) {
      Element digest = new Element("digest");
      // File segment digested, by position in token stream
      Element fileSegment = new Element("fileSegment");
      if (fileSegment.getAttributes().size() > 0) {
        digest.addContent(fileSegment);
      }
      // Base64 encoded digest
      Element base64 = new Element("base64Encoding");
      if (base64.getAttributes().size() > 0) {
        digest.addContent(base64);
      }
      // Compression setting for digest
      Element compression = new Element("compression");
      if (compression.getAttributes().size() > 0) {
        digest.addContent(compression);
      }

      unknownTokens = new Vector<String>();

      // Should be compressed and base64 encoded (Can be reduced by using
      // Bloom filter of discarded tokens from Dictionary, if included)
      if (unknownTokens != null && unknownTokens.size() > 0) {
        Element unknown = new Element("unknownTokens");
        for (String unknownToken : unknownTokens) {
          Element token = new Element("token");
          token.setAttribute("name", unknownToken);
          unknown.addContent(token);
        }
        digest.addContent(unknown);
      }
      fingerprint.addContent(digest);
    }

    // Dictionary
    try {
      File xml = new File(dictionary.getDictionaryFilename());
      SAXBuilder builder = new SAXBuilder();
      Document doc = (Document) builder.build(xml);

      if (showDictionary) {
        // tokens are ordered alphabetically (because as far as I can tell it is
        // much
        // easier than ordering by position)
        fingerprint.addContent((Element) doc.getRootElement().clone());
      } else {
        Element dict = new Element("dictionary");
        dict.setAttribute("name", dictionary.getDictionaryFilename());
        dict.setAttribute("GUID", dictionary.getGUID());
        dict.setAttribute("version", String.valueOf(dictionary.getVersion()));
        if (dictionary.getCreatingProgram() != null && dictionary.getCreatingProgram().length() > 0)
          dict.setAttribute("creatingProgram", dictionary.getCreatingProgram());
        if (dictionary.getCreation() != null && dictionary.getCreation().length() > 0)
          dict.setAttribute("creation", dictionary.getCreation());
        if (dictionary.getSource() != null && dictionary.getSource().length() > 0)
          dict.setAttribute("source", dictionary.getSource());
        fingerprint.addContent(dict);
      }

      Document digestDoc = new Document(fingerprint);
      FileOutputStream digest = new FileOutputStream(outputFile);
      Format form = Format.getPrettyFormat();
      form.setEncoding("UTF-8");
      XMLOutputter out = new XMLOutputter(form);
      out.output(digestDoc, digest);
      digest.close();
    } catch (FileNotFoundException e) {
      ExceptionHandler.handleException(e);
    } catch (UnsupportedEncodingException e) {
      ExceptionHandler.handleException(e);
    } catch (JDOMException e) {
      ExceptionHandler.handleException(e);
    } catch (IOException e) {
      ExceptionHandler.handleException(e);
    }
  }

  /**
   * Sets the digest output to generate for this fingerprinter
   * 
   * @param config_file
   *          the string XML filename containing the output specifications
   **/
  @SuppressWarnings("unchecked")
  public void outputFields(String config_file) {
    setTerse();

    try {
      File xml = new File(config_file);
      SAXBuilder builder = new SAXBuilder();
      Document doc = (Document) builder.build(xml);
      Element fingerprinter = doc.getRootElement();

      for (Element output : (List<Element>) fingerprinter.getChildren("output")) {
        if (output.getAttributeValue("type").contains("dataSource"))
          showDataSource = true;
        if (output.getAttributeValue("type").contains("digest"))
          showDigest = true;
        if (output.getAttributeValue("type").contains("dictionary"))
          showDictionary = true;
      }
    } catch (FileNotFoundException e) {
      System.out.println("Output configuration file not found: Setting terse output");
    } catch (UnsupportedEncodingException e) {
      ExceptionHandler.handleException(e);
    } catch (JDOMException e) {
      ExceptionHandler.handleException(e);
    } catch (IOException e) {
      ExceptionHandler.handleException(e);
    }
  }
}
