package edu.georgetown.gucs.matcher;

import java.io.File;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;
import java.util.Vector;

import edu.georgetown.gucs.dictionary.Dictionary;
import edu.georgetown.gucs.fingerprinter.Fingerprinter;
import edu.georgetown.gucs.utility.FileLister;
import edu.georgetown.gucs.utility.ObjectBuilder;

/**
 * Outputs a list of files and their scores, comparing a given fingerprint to a directory of files. Files are organized
 * by scores with higher scores appearing higher in the list. Providing a minimum score and a dictionary (if the
 * fingerprint does not already contain one) is optional.
 * 
 * @author Lindsay Neubauer
 */
public class CompareDirectory {

  private Fingerprinter fingerprint;
  private FingerprintMatcher matcher;
  private FileLister fileLister;
  private int minimum_score;
  private Dictionary dictionary;
  Map<Integer, List<String>> scores;

  /**
   * Constructor that sets the matcher, fingerprint file and directory to use for this comparison
   * 
   * @param match
   *          the string name of the matcher to use for this comparison
   * @param fingerprintFile
   *          the string filename of the fingerprint to compare with this directory
   * @param directory
   *          the string directory path to compare with this fingerprint
   */
  public CompareDirectory(String match, String fingerprintFile, String directory) {
    this(match, fingerprintFile, directory, 0, null);
  }

  /**
   * Constructor that sets the matcher, fingerprint file, directory and minimum score to use for this comparison
   * 
   * @param match
   *          the string name of the matcher to use for this comparison
   * @param fingerprintFile
   *          the string filename of the fingerprint to compare with this directory
   * @param directory
   *          the string directory path to compare with this fingerprint
   * @param minimumScore
   *          the <code>int</code> minimum score needed to output a file for this comparison
   */
  public CompareDirectory(String match, String fingerprintFile, String directory, int minimumScore) {
    this(match, fingerprintFile, directory, minimumScore, null);
  }

  /**
   * Constructor that sets the matcher, fingerprint file, directory, minimum score and dictionary to use for this
   * comparison
   * 
   * @param match
   *          the string name of the matcher to use for this comparison
   * @param fingerprintFile
   *          the string filename of the fingerprint to compare with this directory
   * @param directory
   *          the string directory path to compare with this fingerprint
   * @param minimumScore
   *          the <code>int</code> minimum score needed to output a file for this comparison
   * @param dictionary
   *          the string filename of the dictionary to use for this comparison; this dictionary will only be used if the
   *          fingerprint file does not already contain a full dictionary
   */
  public CompareDirectory(String match, String fingerprintFile, String directory, int minimumScore, String dictionary) {
    matcher = (FingerprintMatcher) ObjectBuilder.buildObject(match);
    scores = new TreeMap<Integer, List<String>>();
    fileLister = new FileLister();
    fileLister.recursiveLoadDirectory(directory);

    minimum_score = (minimumScore < 0 || minimumScore > 99) ? 0 : minimumScore;

    fingerprint = new Fingerprinter();
    fingerprint.loadFingerprintXML(fingerprintFile, dictionary);
    this.dictionary = fingerprint.getDictionary();
  }

  /**
   * Compute a fingerprint for each file in this directory and compare it to this fingerprint using this matcher; if the
   * matcher returns boolean values, true values are given a score of 99 and false values are given a score of 0
   */
  public void computeFingerprints() {
    Fingerprinter fingerprinter = (Fingerprinter) ObjectBuilder.buildObject(fingerprint.getFingerprintName());
    fingerprinter.setDictionary(dictionary);
    String matcherName = matcher.getClass().getSimpleName();
    int score;

    for (File file : fileLister) {
      String path = file.getPath();
      if (matcherName.equals("HashFingerprintMatcher") || matcherName.equals("ExactFingerprintMatcher")) {
        boolean match = matcher.match(fingerprint.getBase64Fingerprint(), fingerprinter.computeFingerprintXML(path));
        if (match)
          score = 99;
        else
          score = 0;
      } else {
        score = matcher.getScoreXML(fingerprint.getBase64Fingerprint(), fingerprinter.computeFingerprintXML(path));
      }

      List<String> updatedScores;
      if (scores.containsKey(score))
        updatedScores = scores.get(score);
      else
        updatedScores = new Vector<String>();
      updatedScores.add(path);
      scores.put(score, updatedScores);
    }
  }

  /**
   * Print all the files and their similarity scores; if a minimum score is provided, only output those files with
   * scores greater than or equal to the minimum
   */
  public void printScores() {
    boolean printedSomething = false;
    for (int i = 99; i >= minimum_score; i--) {
      if (scores.containsKey(i)) {
        for (String score : scores.get(i)) {
          System.out.println(score + " (score = " + i + ")");
        }
        System.out.println();
        printedSomething = true;
      }
    }
    if (!printedSomething)
      System.out.println("CompareDictionary: There are no files in the directory with a score higher than "
          + minimum_score);
  }

  /**
   * Outputs a list of files and their scores, comparing a given fingerprint to a directory of files. Files are
   * organized by scores with higher scores appearing higher in the list. Providing a minimum score and a dictionary (if
   * the fingerprint does not already contain one) is optional.
   * 
   * @param args
   *          array of string command line arguments<br>
   *          <code>args[0]</code> the name of the matcher to use for this comparison<br>
   *          <code>args[1]</code> the filename of the fingerprint to compare with this directory<br>
   *          <code>args[2]</code> the directory path to compare with this fingerprint<br>
   *          <code>args[3]</code> the optional integer minimum score needed to output a file for this comparison<br>
   *          <code>args[4]</code> the optional filename of the dictionary to use for this comparison; only used if the
   *          fingerprint file does not contain a full dictionary
   */
  public static void main(String args[]) {
    String matcher = null;
    String fingerprint = null;
    String directory = null;
    String dictionary = null;
    int min_score = 0;

    try {
      matcher = args[0];
      fingerprint = args[1];
      directory = args[2];
      if (args.length > 3)
        min_score = Integer.valueOf(args[3]);
      if (args.length > 4)
        dictionary = args[4];
    } catch (ArrayIndexOutOfBoundsException e) {
      System.err
          .println("Usage: java CompareDirectory <matcher name> <XML fingeprint file> <directory> <optional: minimum score> "
              + "<optional: dictionaryFilename>");
      System.err
          .println("Outputs a list of files and their scores, comparing a given fingerprint to a directory of files");
      System.exit(-1);
    }

    CompareDirectory compare = new CompareDirectory(matcher, fingerprint, directory, min_score, dictionary);
    compare.computeFingerprints();
    compare.printScores();
  }
}
