package edu.georgetown.gucs.tokenizers;

import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.Dictionary;
import java.util.zip.GZIPInputStream;

import edu.georgetown.gucs.utility.ExceptionHandler;
import edu.georgetown.gucs.utility.Pair;

/**
 * Splits contents of a gzipped text file into tokens based on whitespace or by line. Used for {@linkplain Dictionary}
 * creation.
 * 
 * @author Clay Shields
 * @author Lindsay Neubauer
 */
public class GzippedFileTokenizer extends FileTokenizer {

  /**
   * Constructor that sets the token creation mode to split based on whitespace
   */
  public GzippedFileTokenizer() {
    mode = "tokens";
  }

  /**
   * Constructor that sets the token creation mode. If set to "lines" the file is split by line; if set to "tokens" the
   * file is split based on whitespace.
   * 
   * @param mode
   *          the string mode to split tokens based on whitespace or by line
   */
  public GzippedFileTokenizer(String mode) {
    this.mode = mode.trim();
  }

  /**
   * Splits a gzipped text document into tokens.
   */
  public void tokenize(String filename) {
    if (!filename.contains(".gz"))
      System.err.println("GzippedFileTokenizer: filename does not contain gzip file extension");
    readFile(filename);
  }

  /**
   * Splits a gzipped text document into tokens.
   */
  protected void readFile(String filename) {
    try {
      // The encoding here needs to match the encoding of the file that is being
      // read
      // This is assuming the file is UTF-8 encoded (otherwise, it will spit out
      // garbage)
      BufferedReader buffer_reader = new BufferedReader(new InputStreamReader(new GZIPInputStream(new FileInputStream(
          filename)), "UTF-8"));

      tokenVector.clear();
      positions.clear();

      StringBuffer token = new StringBuffer("");
      int character;
      int byte_start = 0;
      int byte_end = 0;

      while ((character = buffer_reader.read()) != -1) {
        if (mode.equalsIgnoreCase("lines")) {
          if ('\n' == character || '\r' == character || '\f' == character) {
            if (token.toString().trim().length() > 0) {
              tokenVector.add(token.toString().trim());
              positions.add(new Pair<Integer, Integer>(byte_start, byte_end));
              byte_start = byte_end;
            }
            byte_end += Character.charCount(character);
            token.delete(0, token.length());
          } else {
            token.append((char) character);
            byte_end += Character.charCount(character);
          }
        } else {
          if (Character.isWhitespace(character)) {
            if (token.toString().trim().length() > 0) {
              tokenVector.add(token.toString().trim());
              positions.add(new Pair<Integer, Integer>(byte_start, byte_end));
              byte_start = byte_end;
            }
            byte_end += Character.charCount(character);
            token.delete(0, token.length());
          } else {
            token.append((char) character);
            byte_end += Character.charCount(character);
          }
        }
      }
      if (token.toString().trim().length() > 0) {
        tokenVector.add(token.toString().trim());
        positions.add(new Pair<Integer, Integer>(byte_start, byte_end));
      }
      buffer_reader.close();
    } catch (IOException e) {
      ExceptionHandler.handleException(e);
    }
  }

  /**
   * Prints to the screen a portion of the given gzipped text document
   */
  public void checkIndexing(String filename, int start_index, int end_index) {
    try {
      if (!filename.contains(".gz"))
        System.err.println("GzippedFileTokenizer: filename does not contain gzip file extension");
      BufferedReader buffer_reader = new BufferedReader(new InputStreamReader(new GZIPInputStream(new FileInputStream(
          filename)), "UTF-8"));
      long byte_position = buffer_reader.skip(start_index);
      StringBuffer str = new StringBuffer("");
      while (byte_position < end_index) {
        int character = buffer_reader.read();
        str.append((char) character);
        byte_position += Character.charCount(character);
      }
      System.out.println(str.toString());
      buffer_reader.close();
    } catch (IOException e) {
      ExceptionHandler.handleException(e);
    }
  }

  /**
   * Tokenizes a gzipped text file and prints the resulting tokens to the the screen
   * 
   * @param args
   *          array of string command line arguments<br>
   *          <code>args[0]</code> the filename of the English language gzipped document to split into tokens
   */
  public static void main(String args[]) {
    String filename = "";
    try {
      filename = args[0];
    } catch (ArrayIndexOutOfBoundsException e) {
      System.err.println("Usage: java -jar GzippedFileTokenizer.jar <file name>");
      System.err.println("Tokenizes a gzipped text file and prints the resulting tokens to the the screen");
      System.exit(-1);
    }
    if (!filename.endsWith(".gz"))
      System.err.println("GzippedFileTokenizer: filename does not contain gzip file extension");
    GzippedFileTokenizer test = new GzippedFileTokenizer();
    test.tokenize(filename);
    printTokens(test.iterator());
  }
}
