package edu.georgetown.gucs.tokenizers;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Locale;
import java.util.Map;

import edu.georgetown.gucs.utility.Pair;

/**
 * Eliminates tokens specified in the given stop words document; Assumes this document contains lower case,
 * non-porterized English word
 * 
 * @author Clay Shields
 * @author Lindsay Neubauer
 * 
 */
public class StopWordRemoverTokenizer extends Tokenizer {

  private Map<String, Object> stopwords = new HashMap<String, Object>();

  /**
   * Constructor that loads stop words to use when <code>tokenize</code> method is called;
   * 
   * @param stopwordFilename
   *          the string filename of the document containing the stop words to use; assumes the document contains lower
   *          case, non-porterized English word
   */
  public StopWordRemoverTokenizer(String stopwordFilename) {
    stopwords = new HashMap<String, Object>();
    loadStopWords(stopwordFilename);
  }

  /**
   * Loads words to eliminate when <code>tokenize</code> method is called
   * 
   * @param stopwordFilename
   *          the string filename of the document containing the stop words to use; assumes the document contains lower
   *          case, non-porterized English word
   */
  public void loadStopWords(String stopwordFilename) {
    FileTokenizer stops = new FileTokenizer();
    stops.tokenize(stopwordFilename);
    Iterator<String> iterator = stops.iterator();
    while (iterator.hasNext()) {
      String token = iterator.next();
      stopwords.put(token, null);
    }
  }

  /**
   * Eliminates tokens specified in the given stop words document
   */
  public void tokenize(Iterator<String> iterator) {
    tokenVector.clear();
    while (iterator.hasNext()) {
      String token = iterator.next();
      String lctoken = token.toLowerCase(Locale.ENGLISH);
      if (!stopwords.containsKey(lctoken)) {
        tokenVector.add(token);
      }
    }
  }

  /**
   * Eliminates tokens specified in the given stop words document
   */
  public void tokenize(Iterator<String> tokensIterator, Iterator<Pair<Integer, Integer>> positionsIterator) {
    tokenVector.clear();
    positions.clear();
    while (tokensIterator.hasNext() && positionsIterator.hasNext()) {
      String token = tokensIterator.next().trim();
      Pair<Integer, Integer> pos = positionsIterator.next();
      String lctoken = token.toLowerCase(Locale.ENGLISH);
      if (!stopwords.containsKey(lctoken)) {
        tokenVector.add(token);
        positions.add(pos);
      }
    }
  }
}
