package edu.georgetown.gucs.utility;

import java.nio.ByteBuffer;
import java.util.List;
import java.util.Locale;

/**
 * Creates a BitSet to replace {@linkplain java.util.BitSet}. The primary difference between the two are that most set
 * operators return new sets as opposed to ORing and ANDing "in place".
 * 
 * @see <a href="http://antlr.sourcearchive.com/documentation/2.7.2-6/BitSet_8java-source.html">BitSet.java</a>
 * 
 * @author Terence Parr
 * @author <a href="mailto:pete@yamuna.demon.co.uk">Pete Wells</a>
 * 
 * @author Clay Shields
 * @author Lindsay Neubauer
 */
public class AntlrBitSet implements Cloneable {

  /*
   * Cannot contain a BitSet because there is no way to access the internal bits (which are needed for speed) and,
   * because it is final, cannot use a subclass to add functionality. Consider defining set degree. Without access to
   * the bits, must call a method n times to test the ith bit...ack!
   */

  protected final static int BITS = 64; // number of bits / long
  protected final static int NIBBLE = 4;
  protected final static int LOG_BITS = 6; // 2^6 == 64

  /**
   * A precomputed mod mask. For powers of two, the mod operation (i mod nbits) is the same as (i & (nbits-1))
   */
  protected final static int MOD_MASK = BITS - 1;

  /** The actual data bits */
  protected long bits[];

  /**
   * Constructs a bitset of size one word (64 bits)
   */
  public AntlrBitSet() {
    this(BITS);
  }

  /**
   * Constructs a clone of the static array of longs
   */
  public AntlrBitSet(long[] bits_) {
    bits = bits_.clone();
  }

  /**
   * Constructs a bitset of given size
   * 
   * @param nbits
   *          the <code>int<code> size of the bitset in bits
   */
  public AntlrBitSet(int nbits) {
    bits = new long[((nbits - 1) >> LOG_BITS) + 1];
  }

  /**
   * Adds an element to this set using OR operator; set grows as necessary to accommodate
   * 
   * @param el
   *          the <code>int</code> element to add
   */
  public void add(int el) {
    int n = wordNumber(el);
    if (n >= bits.length) {
      growToInclude(el);
    }
    bits[n] |= bitMask(el);
  }

  /**
   * Performs AND operation
   * 
   * @param a
   *          the antlrBitSet to perform AND operation with
   * 
   * @return the resulting antlrBitSet
   */
  public AntlrBitSet and(AntlrBitSet a) {
    AntlrBitSet s = (AntlrBitSet) this.clone();
    s.andInPlace(a);
    return s;
  }

  /**
   * Performs AND operation
   * 
   * @param a
   *          the antlrBitSet to perform AND operation with
   */
  public void andInPlace(AntlrBitSet a) {
    int min = Math.min(bits.length, a.bits.length);
    for (int i = min - 1; i >= 0; i--) {
      bits[i] &= a.bits[i];
    }
    // clear all bits in this not present in a (if this bigger than a).
    for (int i = min; i < bits.length; i++) {
      bits[i] = 0;
    }
  }

  /**
   * TODO
   * 
   * @param bitNumber
   * @return the <code>long</code>
   */
  private final static long bitMask(int bitNumber) {
    int bitPosition = bitNumber & MOD_MASK; // bitNumber mod BITS
    return 1L << bitPosition;
  }

  /**
   * Clears all elements in this bitset
   */
  public void clear() {
    for (int i = bits.length - 1; i >= 0; i--) {
      bits[i] = 0;
    }
  }

  /**
   * Clears the given element in this bitset
   * 
   * @param el
   *          the element to clear
   */
  public void clear(int el) {
    int n = wordNumber(el);
    if (n >= bits.length) { // grow as necessary to accommodate
      growToInclude(el);
    }
    bits[n] &= ~bitMask(el);
  }

  /**
   * Creates and returns a copy of this antlrBitSet
   */
  public AntlrBitSet clone() {
    AntlrBitSet s = null;
    try {
      s = (AntlrBitSet) super.clone();
      s.bits = new long[bits.length];
      System.arraycopy(bits, 0, s.bits, 0, bits.length);
    } catch (CloneNotSupportedException e) {
      ExceptionHandler.handleException(new InternalError());
    }
    return s;
  }

  /**
   * Returns the degree of this bitset
   * 
   * @return the <code>int</code> degree
   */
  public int degree() {
    int deg = 0;
    for (int i = bits.length - 1; i >= 0; i--) {
      long word = bits[i];
      if (word != 0L) {
        for (int bit = BITS - 1; bit >= 0; bit--) {
          if ((word & (1L << bit)) != 0) {
            deg++;
          }
        }
      }
    }
    return deg;
  }

  /**
   * Returns true if the <code>obj</code> is a bit set that contains exactly the same elements as this bit set,
   * otherwise false.
   * <p>
   * Code "inherited" from java.util.BitSet
   * 
   * @param obj
   *          the object to compare to
   * 
   * @return the <code>boolean</code> whether obj equals this bit set
   */
  public boolean equals(Object obj) {
    if ((obj != null) && (obj instanceof AntlrBitSet)) {
      AntlrBitSet set = (AntlrBitSet) obj;

      int n = Math.min(bits.length, set.bits.length);
      for (int i = n; i-- > 0;) {
        if (bits[i] != set.bits[i]) {
          return false;
        }
      }
      if (bits.length > n) {
        for (int i = bits.length; i-- > n;) {
          if (bits[i] != 0) {
            return false;
          }
        }
      } else if (set.bits.length > n) {
        for (int i = set.bits.length; i-- > n;) {
          if (set.bits[i] != 0) {
            return false;
          }
        }
      }
      return true;
    }
    return false;
  }

  /**
   * Returns a hash code value for this bit set. The hash code of two bit sets containing the same integers is
   * identical. The algorithm used to compute it is as follows:
   * <p>
   * Suppose the bits in the BitSet were to be stored in an array of long integers called <code>bits</code>, in such a
   * manner that bit <code>k</code> is set in the BitSet (for non-negative values of <code>k</code>) if and only if
   * <p>
   * <code>((k/64) &lt; bits.length)
   * && ((bits[k/64] & (1L &lt;&lt; (bit % 64))) != 0)
   * </code>
   * <p>
   * Then the following definition of the hashCode method would be a correct implementation of the actual algorithm:
   * <p>
   * Code "inherited" from java.util.BitSet
   */
  public int hashCode() {
    long h = 1234;
    for (int i = bits.length; i > 0;)
      h ^= i * bits[--i];
    return (int) ((h >> 32) ^ h);
  }

  /**
   * Grows the set to a larger number of bits
   * 
   * @param bit
   *          the element that must fit in set
   */
  public void growToInclude(int bit) {
    int newSize = Math.max(bits.length << 1, numWordsToHold(bit));
    long newbits[] = new long[newSize];
    System.arraycopy(bits, 0, newbits, 0, bits.length);
    bits = newbits;
  }

  /**
   * TODO
   * 
   * @param el
   * @return the <code>boolean</code>
   */
  public boolean member(int el) {
    int n = wordNumber(el);
    if (n >= bits.length)
      return false;
    return (bits[n] & bitMask(el)) != 0;
  }

  /**
   * Returns whether this bitset is empty
   * 
   * @return the <code>boolean</code> whether the bitset contains all zeros
   */
  public boolean nil() {
    for (int i = bits.length - 1; i >= 0; i--) {
      if (bits[i] != 0)
        return false;
    }
    return true;
  }

  /**
   * Provides the bitwise complement of this bitset
   * 
   * @return the bitset containing the bitwise complement of this bitset
   */
  public AntlrBitSet not() {
    AntlrBitSet s = (AntlrBitSet) this.clone();
    s.notInPlace();
    return s;
  }

  /**
   * Performs a bitwise complement to invert the bits in this bitset
   */
  public void notInPlace() {
    for (int i = bits.length - 1; i >= 0; i--) {
      bits[i] = ~bits[i];
    }
  }

  /**
   * Performs a bitwise complement to invert the bits from zero to maxBit
   * 
   * @param maxBit
   *          the maximum bit to complement
   */
  public void notInPlace(int maxBit) {
    notInPlace(0, maxBit);
  }

  /**
   * Performs a bitwise complement to invert the bits from minBit to maxBit
   * 
   * @param minBit
   *          the minumum bit to complement
   * @param maxBit
   *          the maximum bit to complement
   */
  public void notInPlace(int minBit, int maxBit) {
    // make sure that we have room for maxBit
    growToInclude(maxBit);
    for (int i = minBit; i <= maxBit; i++) {
      int n = wordNumber(i);
      bits[n] ^= bitMask(i);
    }
  }

  /**
   * TODO
   * 
   * @param el
   * @return the <code>int</code>
   */
  private final int numWordsToHold(int el) {
    return (el >> LOG_BITS) + 1;
  }

  /**
   * Creates a antlrBitSet with the given element
   * 
   * @param el
   *          the <code>int<code> element to add
   * 
   * @return the resulting antlrBitSet
   */
  public static AntlrBitSet of(int el) {
    AntlrBitSet s = new AntlrBitSet(el + 1);
    s.add(el);
    return s;
  }

  /**
   * Performs OR operation
   * 
   * @param a
   *          the antlrBitSet to perform OR operation with
   * 
   * @return the resulting antlrBitSet
   */
  public AntlrBitSet or(AntlrBitSet a) {
    AntlrBitSet s = (AntlrBitSet) this.clone();
    s.orInPlace(a);
    return s;
  }

  /**
   * Performs OR operation
   * 
   * @param a
   *          the antlrBitSet to perform OR operation with
   */
  public void orInPlace(AntlrBitSet a) {
    // If this is smaller than a, grow this first
    if (a.bits.length > bits.length) {
      setSize(a.bits.length);
    }
    int min = Math.min(bits.length, a.bits.length);
    for (int i = min - 1; i >= 0; i--) {
      bits[i] |= a.bits[i];
    }
  }

  /**
   * Removes the given element from this set
   * 
   * @param el
   *          the <code>int<code> element to remove
   */
  public void remove(int el) {
    int n = wordNumber(el);
    if (n >= bits.length) {
      growToInclude(el);
    }
    bits[n] &= ~bitMask(el);
  }

  /**
   * Sets the size of a set
   * 
   * @param nwords
   *          the <code>int</code> number of words the new set should be
   */
  private void setSize(int nwords) {
    long newbits[] = new long[nwords];
    int n = Math.min(nwords, bits.length);
    System.arraycopy(bits, 0, newbits, 0, n);
    bits = newbits;
  }

  /**
   * TODO
   * 
   * @return the <code>int</code>
   */
  public int size() {
    return bits.length << LOG_BITS; // num words * bits per word
  }

  /**
   * Returns the amount of space being used by the bits array (not the number of member bits on)
   */
  public int lengthInLongWords() {
    return bits.length;
  }

  /**
   * Returns whether this bitset is contained within a
   * 
   * @param a
   *          the antlrBitSet to check
   * 
   * @return the <code>boolean</code> whether this bitset is in the provided antlrBitSet
   */
  public boolean subset(AntlrBitSet a) {
    if (a == null)
      return false;
    return this.and(a).equals(this);
  }

  /**
   * Subtracts the elements of the given antlrBitSet from this bitset in-place (turn off all bits of this bitset that
   * are in the given antlrBitSet)
   * 
   * @param a
   *          the antlrBitSet to subtract the element from
   */
  public void subtractInPlace(AntlrBitSet a) {
    if (a == null)
      return;
    // for all words of 'a', turn off corresponding bits of 'this'
    for (int i = 0; i < bits.length && i < a.bits.length; i++) {
      bits[i] &= ~a.bits[i];
    }
  }

  /**
   * TODO
   * 
   * @return the array of integers
   */
  public int[] toArray() {
    int[] elems = new int[degree()];
    int en = 0;
    for (int i = 0; i < (bits.length << LOG_BITS); i++) {
      if (member(i)) {
        elems[en++] = i;
      }
    }
    return elems;
  }

  /**
   * Creates and returns a copy of this antlrBitSet's bitset
   * 
   * @return the <code>long</code> array containing a clone of this bitset
   */
  public long[] toPackedArray() {
    return bits.clone();
  }

  /**
   * Transform a bit set into a string separated by commas by formatting each element as an integer
   * 
   * @return A comma-separated list of values
   */
  public String toString() {
    return toString(",");
  }

  /**
   * Transform a bit set into a string by formatting each element as an integer
   * 
   * @param separator
   *          the string to put in between elements
   * 
   * @return A comma-separated list of values
   */
  public String toString(String separator) {

    StringBuffer buffer = new StringBuffer();
    for (int i = 0; i < (bits.length << LOG_BITS); i++) {
      if (member(i)) {
        if (buffer.length() > 0) {
          buffer.append(separator);
        }
        buffer.append(i);
      }
    }
    return buffer.toString();
  }

  /**
   * Creates a string representation where instead of integer elements, the ith element of a list of strings is
   * displayed
   * 
   * @param separator
   *          the string to put in between elements
   * @param vocabulary
   *          the list of display elements
   * 
   * @return A comma-separated list of character constants
   */
  public String toString(String separator, List<String> vocabulary) {
    if (vocabulary == null) {
      return toString(separator);
    }
    StringBuffer buffer = new StringBuffer();
    for (int i = 0; i < (bits.length << LOG_BITS); i++) {
      if (member(i)) {
        if (buffer.length() > 0) {
          buffer.append(separator);
        }
        if (i >= vocabulary.size()) {
          buffer.append("<bad element " + i + ">");
        } else if (vocabulary.get(i) == null) {
          buffer.append("<" + i + ">");
        } else {
          buffer.append(vocabulary.get(i));
        }
      }
    }
    return buffer.toString();
  }

  /**
   * Dumps a comma-separated list of the words making up the bit set; Splits each 64 bit number into two more manageable
   * 32 bit numbers; Generates a comma-separated list of C++-like unsigned long constants
   */
  public String toStringOfHalfWords() {
    StringBuffer buffer = new StringBuffer();
    for (int i = 0; i < bits.length; i++) {
      if (i != 0)
        buffer.append(", ");
      long tmp = bits[i];
      tmp &= 0xFFFFFFFFL;
      buffer.append(tmp + "UL");
      buffer.append(", ");
      tmp = bits[i] >>> 32;
      tmp &= 0xFFFFFFFFL;
      buffer.append(tmp + "UL");
    }
    return buffer.toString();
  }

  /**
   * Dumps a comma-separated list of the words making up the bit set; Generates a comma-separated list of Java-like long
   * int constants
   */
  public String toStringOfWords() {
    StringBuffer buffer = new StringBuffer();
    for (int i = 0; i < bits.length; i++) {
      if (i != 0)
        buffer.append(", ");
      buffer.append(bits[i] + "L");
    }
    return buffer.toString();
  }

  /**
   * TODO
   * 
   * @param bit
   * @return the <code>int</code>
   */
  private final static int wordNumber(int bit) {
    return bit >> LOG_BITS; // bit / BITS
  }

  /* ********************************************** */
  /* ************** SDTEXT ADDITIONS ************** */
  /* ********************************************** */

  /**
   * Converts this bitset to a hexadecimal string
   * 
   * @return the hexadecimal string
   */
  public String toHex() {
    StringBuffer hexBuffer = new StringBuffer();
    for (int i = 0; i < bits.length; i++) {
      String hex = Long.toHexString(bits[i]).toUpperCase(Locale.ENGLISH);
      StringBuffer zeros = new StringBuffer();
      while ((zeros.length() + hex.length()) % 16 != 0) {
        zeros.append("0");
      }
      String thisLong = zeros.toString() + hex;
      hexBuffer.append(thisLong);
    }
    return hexBuffer.toString();
  }

  /**
   * Converts bytes to longs
   * 
   * @param bytes
   *          the array of bytes to convert
   * 
   * @return the <code>long</code> array
   */
  private long[] bytesToLongs(byte[] bytes) {
    int bitLen = bytes.length / 8;
    if ((bytes.length % 8) != 0) {
      bitLen++;
      System.err.println("The bytes given to AntlrBitSet were not a multiple of 8");
      System.err.println("We may not handle this well");
    }
    long[] tempBits = new long[bitLen];
    int byteBase = 0;
    for (int longIndex = 0; longIndex < bitLen; longIndex++) {
      long l = 0;
      for (int i = 0; i < 8; i++) {
        l <<= 8; // this is the same as l = l << 8;
        l |= bytes[byteBase] & 0xFF;
        byteBase++;
      }
      tempBits[longIndex] = l;
    }
    return tempBits;
  }

  /**
   * Constructor that converts and sets this bitset
   * 
   * @param bytes
   *          the array of bytes to convert and set
   */
  public AntlrBitSet(byte[] bytes) {
    bits = bytesToLongs(bytes);
  }

  /**
   * Converts this bitset to an array of bytes
   * 
   * @return the array of bytes
   */
  public byte[] toBytes() {
    byte[] bytes = new byte[bits.length * 8];
    int byteBase = 0;
    for (int i = 0; i < bits.length; i++) {
      byte b[] = new byte[8];
      ByteBuffer buf = ByteBuffer.wrap(b);
      buf.putLong(bits[i]);
      for (int j = 0; j < 8; j++) {
        bytes[byteBase] = b[j];
        byteBase++;
      }
    }
    return bytes;
  }

  /**
   * Returns the number of bits that are the same between the antlrBitSet and the provided byte[] bitset
   * 
   * @param bytes
   *          the array of bytes to compare
   * 
   * @return the <code>int</code> number of different bits; returns 0 if the lengths of the two bitsets are different
   */
  public int degreeOfSimilarity(byte[] bytes) {
    long[] otherLongs = bytesToLongs(bytes);

    if (otherLongs.length != bits.length) {
      System.err.println("Cant use degreeOfSimilarity with different sizes, returning zero");
      return 0;
    }

    // Do the and
    for (int i = bits.length - 1; i >= 0; i--) {
      otherLongs[i] &= bits[i];
    }

    // Count the degree: cut and paste of degree code, hacky!
    int deg = 0;
    for (int i = otherLongs.length - 1; i >= 0; i--) {
      long word = otherLongs[i];
      if (word != 0L) {
        for (int bit = BITS - 1; bit >= 0; bit--) {
          if ((word & (1L << bit)) != 0) {
            deg++;
          }
        }
      }
    }
    return deg;
  }

  /**
   * Returns the number of bits that differ between the AntlrBitSet and the provided byte[] bitset
   * 
   * @param bytes
   *          the array of bytes to compare
   * 
   * @return the <code>int</code> number of different bits; returns 0 if the lengths of the two bitsets are different
   */
  public int degreeOfDifference(byte[] bytes) {
    long[] otherLongs = bytesToLongs(bytes);

    if (otherLongs.length != bits.length) {
      System.err.println("Cant use degreeOfDifference with differentsizes, returning zero");
      return 0;
    }

    // Do the XOR
    for (int i = bits.length - 1; i >= 0; i--) {
      otherLongs[i] ^= bits[i];
    }

    // Count the degree: cut and paste of degree code, hacky!
    int deg = 0;
    for (int i = otherLongs.length - 1; i >= 0; i--) {
      long word = otherLongs[i];
      if (word != 0L) {
        for (int bit = BITS - 1; bit >= 0; bit--) {
          if ((word & (1L << bit)) != 0) {
            deg++;
          }
        }
      }
    }
    return deg;
  }

  /**
   * Performs XOR operation
   * 
   * @param a
   *          the antlrBitSet to perform XOR operation with
   */
  public void xorInPlace(AntlrBitSet a) {
    // If this is smaller than a, grow this first
    if (a.bits.length > bits.length) {
      setSize(a.bits.length);
    }
    int min = Math.min(bits.length, a.bits.length);
    for (int i = min - 1; i >= 0; i--) {
      bits[i] ^= a.bits[i];
    }
  }
}