package edu.georgetown.gucs.utility;

import java.io.File;
import java.io.Serializable;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Vector;

/**
 * Creates a list of files in a directory and allows iteration over all and parts of this list
 * 
 * @author Clay Shields
 * @author Lindsay Neubauer
 */
public class FileLister implements Serializable, Iterable<File> {

  private static final long serialVersionUID = 42L; // arbitrary
                                                    // serialVersionUID
  private String directory;
  private List<File> fileList = new Vector<File>();
  private List<File> newFileList = new Vector<File>();
  private Map<String, Integer> fileTypes = new HashMap<String, Integer>();

  /**
   * Default constructor
   */
  public FileLister() {
  }

  /**
   * Constructor that loads all the files from the given directory
   * 
   * @param directory
   *          the string path to the directory to load
   */
  public FileLister(String directory) {
    loadDirectory(directory);
  }

  /**
   * Provides the counts for each file extension that exists in this list of files
   * 
   * @return the hashMap of file extensions and their integer counts in this list of files
   */
  public Map<String, Integer> getFileTypes() {
    return fileTypes;
  }

  /**
   * Provides an iterator over the list of files
   * 
   * @return the file iterator over the list of files
   */
  @Override
  public Iterator<File> iterator() {
    return fileList.iterator();
  }

  /**
   * Provides the size of the initial list of files
   * 
   * @return the integer size of the initial list of files
   */
  public int getNumberOfFiles() {
    return fileList.size();
  }

  /**
   * Provides the path to the directory used for the initial list of files
   * 
   * @return the string path to the directory used to initially create this list of files
   */
  public String getDirectory() {
    return directory;
  }

  /**
   * Provides an iterator over the updated list of files; need to call <code>newCountIterator</code> or
   * <code>newPercentIterator</code> to update the list of files
   * 
   * @return the file iterator over the updated list of files
   */
  public Iterator<File> getNewListIterator() {
    return newFileList.iterator();
  }

  /**
   * Loads all the files from the given directory into the initial list of files
   * 
   * @param directory
   *          the string path to the directory to load
   */
  public void loadDirectory(String directory) {

    // Get the directory listing
    this.directory = (new File(directory)).getAbsolutePath();
    File root = new File(directory);
    File[] files = root.listFiles();

    // Add all readable files to the allFiles array
    for (int i = 0; i < files.length; i++) {
      if (files[i].isFile() && files[i].canRead()) {
        fileList.add(files[i]);
        updateFileTypes(files[i].getName());
      }
    }
  }

  /**
   * Creates a new list of randomly chosen files of the given size from this fileLister
   * 
   * @param count
   *          the <code>int</code> number of the files to include
   * 
   * @return the iterator over the files used to create this fileLister
   */
  public Iterator<File> createNewListFromCount(int count) {
    // Shuffle the current fileList to get a random collection of files
    Collections.shuffle(fileList);

    newFileList = new Vector<File>();
    if (count <= fileList.size()) {
      for (int i = 0; i < count; i++) {
        newFileList.add(fileList.remove(0));
      }
    } else {
      System.err.println("WARNING: FileLister: Asked for more elements(" + count + ") than exist(" + fileList.size()
          + ")");
      System.err.println("         Creating a list with size " + fileList.size());
      newFileList = fileList;
    }
    return newFileList.iterator();
  }

  /**
   * Creates a new list of randomly chosen files that is a percent of the files from this fileLister
   * 
   * @param percent
   *          the <code>int</code> percent of the files to include
   * 
   * @return the iterator over the files used to create this fileLister
   */
  public Iterator<File> createNewListFromPercent(double percent) {
    return createNewListFromCount((int) (percent * fileList.size()));
  }

  /**
   * Prints each file from the initial list of files to standard output
   */
  public void printFiles() {
    for (File file : this) {
      System.out.println(file.toString());
    }
  }

  /**
   * Recursively loads all the files from the given directory into the initial list of files
   * 
   * @param directory
   *          the string path to the directory to load
   */
  public void recursiveLoadDirectory(String directory) {
    this.directory = directory;
    // Get the directory listing
    File root = new File(directory);
    File[] files = root.listFiles();

    // Add all readable files to the allFiles array
    for (int i = 0; i < files.length; i++) {
      if (files[i].isFile() && files[i].canRead()) {
        fileList.add(files[i]);
        updateFileTypes(files[i].getName());
      } else if (files[i].isDirectory() && files[i].canRead()) {
        // Ignore .svn directories
        if (files[i].toString().indexOf(".svn") < 0) {
          recursiveLoadDirectory(files[i].toString());
          directory = (new File(directory)).getAbsolutePath();
        }
      }
    }
  }

  /**
   * Increments the count of the file extension type for the given filename
   * 
   * @param filename
   *          the string filename of file extension to add
   */
  private void updateFileTypes(String filename) {
    String extension = filename.substring(filename.lastIndexOf('.') + 1).toUpperCase(Locale.ENGLISH);

    if (fileTypes.containsKey(extension)) {
      int count = fileTypes.get(extension);
      fileTypes.put(extension, count + 1);
    } else {
      fileTypes.put(extension, 1);
    }
  }
}
